import numpy as np
from peegy.processing.statistics.definitions import HotellingTSquareTest
from peegy.definitions.channel_definitions import ChannelItem
from peegy.processing.tools.eeg_epoch_operators import w_mean, effective_sampling_size
from peegy.processing.tools.epochs_processing_tools import goertzel
from peegy.tools.units.unit_tools import set_default_unit
import astropy.units as u
import scipy.stats as st
from tqdm import tqdm


def hotelling_t_square_test(samples=np.array([]),
                            weights: np.array = None,
                            channels: np.array([ChannelItem]) = None,
                            **kwargs):
    """
    Compute Hotelling T2 test on samples
    :param samples: np.array with samples to perform the test (time x channels x trials)
    :param weights: if passed, statistics will be computed using weighted statistics
    :param channels: list of ChannelItems use to generate pandas data frame with the label of each channel
    :param kwargs:
    :return: pandas data frame with statistical tests
    """
    samples = set_default_unit(samples, u.uV)
    _test_data = []
    if not samples.size:
        return None
    if weights is None:
        weights = np.ones(samples.shape[1::]) * u.dimensionless_unscaled
    if weights.ndim == 3:
        # we assume that the same weight is use for each row
        weights = weights[0, :, :]

    mean_data = w_mean(samples, weights=weights)
    dof = effective_sampling_size(weights, axis=1)
    data_demeaned = samples - mean_data[..., None]
    with np.errstate(divide='ignore', invalid='ignore'):
        for _ch in tqdm(range(samples.shape[1]), desc="Hotelling-T2"):
            # _cov_mat = np.cov(data_ch.value.squeeze(), aweights=_weights.value.squeeze()) * data_ch.unit ** 2.0
            data_ch = data_demeaned[:, _ch, :]
            mean_ch = mean_data[:, _ch]
            _cov_mat = np.cov(data_ch.value, aweights=weights[_ch, :].value) * data_ch.unit ** 2.0
            if _cov_mat.ndim == 0:
                _t_square = dof[_ch] * mean_ch.dot(1 / _cov_mat).dot(mean_ch.T)
            else:
                _t_square = dof[_ch] * mean_ch.dot(np.linalg.pinv(_cov_mat)).dot(mean_ch.T)
            _t_square = _t_square.value
            n_p = data_ch.shape[0]
            n_n = dof[_ch]
            _f = (n_n - n_p) / (n_p * (n_n - 1.0)) * _t_square
            _d1 = n_p
            _d2 = np.maximum(n_n - n_p, 0)
            _f_95 = st.f.ppf(0.95, _d1, _d2)
            if _d2 > 0:
                p = 1 - st.f.cdf(_f, _d1, _d2)
            else:
                p = 1
                _f_95 = np.inf

            # compute residual noise from circular variance
            _rms = np.sqrt(np.mean(mean_ch ** 2.0))
            _rn = np.sqrt(np.sum(data_ch ** 2.0)) / (data_ch.shape[1] - 1)
            _snr = np.maximum(_f - 1, 0.0 * u.dimensionless_unscaled)
            if channels is not None:
                _channel = channels[_ch].label
            else:
                _channel = str(_ch)

            # _L = ((2 * (self.split_sweep_count[i] - 1) * D * F95)./(
            # self.split_sweep_count(i) * (self.split_sweep_count(i) - 2))) ** 2

            test = HotellingTSquareTest(t_square=_t_square,
                                        df_1=_d1,
                                        df_2=_d2,
                                        n_epochs=samples.shape[2],
                                        f=_f,
                                        p_value=p,
                                        mean_amplitude=_rms,
                                        mean_phase=0 * u.rad,
                                        rn=_rn,
                                        snr=_snr,
                                        snr_db=10 * np.log10(_snr) if _snr > 0.0 else
                                        -np.Inf * u.dimensionless_unscaled,
                                        snr_critic_db=10 * np.log10(_f_95 - 1),
                                        snr_critic=_f_95 - 1,
                                        f_critic=_f_95,
                                        channel=_channel,
                                        **kwargs)
            _test_data.append(test)
    return _test_data


def f_test(f_values: np.array([]) = None,
           df_numerator: float = 5,
           df_noise: np.array([]) = None,
           alpha: float = 0.05
           ) -> (float, float, float, float, float):
    """
    Compute F-Test for input samples.
    The degrees of freedom for signal are from Elberling 1984.
    https://www.tandfonline.com/doi/abs/10.3109/01050398409043059
    :param df_numerator: degrees of freedom of the numerator of the F-test
    :param df_noise: The estimated degrees of freedom of the noise used to estimate the input SNR
    :param f_values: f_values to test
    :param alpha: the statistical level for significance
    :return: pandas data frame with statistical tests
    """

    df_num = df_numerator
    df_den = df_noise
    f_critic = st.f.ppf(1 - alpha, df_num, df_den)
    p_value = 1 - st.f.cdf(f_values, df_num, df_den)

    return f_critic, df_num, df_den, p_value


def phase_locking_value(
        data: np.array = None,
        alpha: float = 0.05,
        weights: np.array = None) -> np.array:
    """
    Compute phase-locking value (PLV) using Rayleigh test
    :param data: time x channels x trials numpy array in the time domain
    :param alpha: float indicating the alpha value for significance
    :param weights: trial by trial weights
    :return: phase-locking value (frequency x channels), z-scores (frequency x channels), z_critic (float), p_values
    (frequency x channels), mean angles (frequency x channels in rads), degrees of freedom, residual_noise estimation
    """
    if weights is None:
        weights = np.ones(data.shape)
    _dof = effective_sampling_size(weights)
    # for the moment we assume that the same weight across time is applied so that it can be factor out from fft
    dof = np.min(_dof, axis=0, keepdims=False).squeeze()[None, :]
    _weights = np.min(weights, axis=0)[None, :, :]
    yfft = np.fft.rfft(data, axis=0)
    spectral_mean = w_mean(yfft, _weights)
    diff = yfft - spectral_mean[:, :, None]
    # compute weighted standard deviation
    noise = np.sqrt(w_mean(np.abs(diff) ** 2, _weights) / dof)
    # scale to match time domain amplitudes
    spectral_amp = 2 * np.abs(spectral_mean) / data.shape[0]
    residual_noise = 2 * np.abs(noise) / data.shape[0]
    # normalize
    norm = np.sqrt(np.abs(yfft * np.conj(yfft)))
    yfft = np.divide(yfft, norm, out=np.zeros_like(yfft), where=norm != 0)
    angles = np.angle(w_mean(yfft, _weights))
    plv = np.abs(w_mean(yfft, _weights))
    z = dof * plv ** 2

    # D.Wilkie. Rayleigh Test for Randomness of Circular Data".Applied Statistics.1983.
    tmp = np.ones(z.shape)
    _idx_tmp = (dof < 50).squeeze()

    if np.any(_idx_tmp):
        _tmp_correction = 1.0 + (2.0 * z - z * z) / (4.0 * dof) - (
                24.0 * z - 132.0 * z ** 2.0 + 76.0 * z ** 3.0 - 9.0 * z ** 4.0) / (288.0 * dof * dof)
        tmp[:, _idx_tmp] = _tmp_correction
    z_crit = -np.log(alpha) - (2 * np.log(alpha) + np.log(alpha) ** 2.0) / (4 * dof)
    p_values = np.exp(-z) * tmp

    return spectral_amp, plv, z, z_crit, p_values, angles, dof, residual_noise


def discrete_phase_locking_value(
        data: np.array = None,
        alpha: float = 0.05,
        fs: float = None,
        frequency: float = None,
        weights: np.array = None) -> np.array:
    """
    Compute spectral amplitude, phase, and  phase-locking value (PLV) using Rayleigh test for a single frequency
    component.
    :param data: time x channels x trials numpy array in the time domain
    :param alpha: float indicating the alpha value for significance
    :param fs: float indicating the sampling rate in Hz
    :param frequency: float indicating the frequency of interest
    :param weights: trial by trial weights
    :return: phase-locking value (frequency x channels), z-scores (frequency x channels), z_critic (float), p_values
    (frequency x channels), mean angles (frequency x channels in rads), degrees of freedom, residual_noise estimation
    """
    if weights is None:
        weights = np.ones((1, data.shape[1], data.shape[2]))
    _dof = effective_sampling_size(weights)
    # for the moment we assume that the same weight across time is applied so that it can be factor out from fft
    dof = np.min(_dof, axis=0, keepdims=False).squeeze()[None, :]

    yfft, exact_frequency = goertzel(data, fs, frequency)
    spectral_amp = np.abs(np.sum(yfft * weights, axis=2) / np.sum(weights, axis=2))
    diff = yfft - spectral_amp[:, :, None]
    noise = np.sqrt(np.var(diff, ddof=1, axis=2) / dof)
    # scale to match time domain amplitudes
    spectral_amp = 2 * spectral_amp / data.shape[0]
    residual_noise = 2 * noise / data.shape[0]
    yfft = yfft / np.abs(yfft)
    angles = np.angle(np.sum(yfft * weights, axis=2) / np.sum(weights, axis=2))
    plv = np.abs(np.sum(yfft * weights, axis=2) / np.sum(weights, axis=2))
    z = dof * plv ** 2
    # D.Wilkie. Rayleigh Test for Randomness of Circular Data".Applied Statistics.1983.
    tmp = np.ones(z.shape)
    _idx_tmp = (dof < 50).squeeze()

    if np.any(_idx_tmp):
        _tmp_correction = 1.0 + (2.0 * z - z * z) / (4.0 * dof) - (24.0 * z - 132.0 * z ** 2.0 +
                                                                   76.0 * z ** 3.0 - 9.0 * z ** 4.0) / (
                                      288.0 * dof * dof)
        tmp[:, _idx_tmp] = _tmp_correction
    z_crit = -np.log(alpha) - (2 * np.log(alpha) + np.log(alpha) ** 2.0) / (4 * dof)
    p_values = np.exp(-z) * tmp

    return spectral_amp, plv, z, z_crit, p_values, angles, dof, residual_noise


def get_discrete_frequency_value(
        data: np.array = None,
        fs: float = None,
        frequency: float = None,
        weights: np.array = None) -> np.array:
    """
    Compute spectral amplitude for a single frequency
    component.
    :param data: time x channels x trials numpy array in the time domain
    :param alpha: float indicating the alpha value for significance
    :param fs: float indicating the sampling rate in Hz
    :param frequency: float indicating the frequency of interest
    :param weights: trial by trial weights
    :return: frequency amplitudes
    """
    if weights is None:
        weights = np.ones(data.shape[1::])
    if weights.ndim == 3:
        # we assume that the same weight is use for each row
        weights = weights[0, :, :]
    dof = effective_sampling_size(weights, axis=1)
    yfft, _exact_freq = goertzel(data, fs, frequency)

    spectral_amp = np.abs(np.sum(yfft * weights, axis=2) / np.sum(weights, axis=1))
    diff = yfft - spectral_amp[:, :, None]
    noise = np.sqrt(np.var(diff, ddof=1, axis=2) / dof)
    # scale to match time domain amplitudes
    spectral_amp = 2 * spectral_amp / data.shape[0]
    residual_noise = 2 * noise / data.shape[0]
    yfft = 2 * yfft / data.shape[0]
    return yfft, spectral_amp, residual_noise


def freq_bin_phase_locking_value(
        yfft: np.array = None,
        alpha: float = 0.05,
        weights: np.array = None) -> np.array:
    """
    Compute phase, and  phase-locking value (PLV) using Rayleigh test for an array of trials from a single frequency
    component.
    :param yfft: fft_bin (complex) x channels x trials numpy array in the time domain
    :param alpha: float indicating the alpha value for significance
    :param frequency: float indicating the frequency of interest
    :param weights: trial by trial weights
    :return: phase-locking value (frequency x channels), z-scores (frequency x channels), z_critic (float), p_values
    (frequency x channels), mean angles (frequency x channels in rads), degrees of freedom, residual_noise estimation
    """
    if weights is None:
        weights = np.ones(yfft.shape[1::])
    if weights.ndim == 3:
        # we assume that the same weight is use for each row
        weights = weights[0, :, :]
    dof = effective_sampling_size(weights, axis=1)
    spectral_mean = np.sum(yfft * weights, axis=2) / np.sum(weights, axis=1)
    spectral_amp = np.abs(spectral_mean)
    # subtract evoked response from all epochs
    noise_epochs = yfft - spectral_mean[:, :, None]
    noise_epochs_amp = np.abs(noise_epochs)
    noise_epochs_ang = np.angle(noise_epochs)
    mean_x = np.sum(noise_epochs_amp * np.cos(noise_epochs_ang) * weights,
                    axis=2,
                    keepdims=True) / np.sum(weights, axis=1, keepdims=True)
    mean_y = np.sum(noise_epochs_amp * np.sin(noise_epochs_ang) * weights,
                    axis=2,
                    keepdims=True) / np.sum(weights, axis=1, keepdims=True)
    var_x = np.sum(((noise_epochs_amp * np.cos(noise_epochs_ang) - mean_x) ** 2.0) * weights,
                   axis=2,
                   keepdims=True) / np.sum(weights, axis=1, keepdims=True)
    var_y = np.sum(((noise_epochs_amp * np.sin(noise_epochs_ang) - mean_y) ** 2.0) * weights,
                   axis=2,
                   keepdims=True) / np.sum(weights, axis=1, keepdims=True)
    # we assumed that variance on x and y axis are independent from each other
    # this encapsulates the variance: sigma_r^2 <= variance <= sigma_r^2 + 4 * sum(w_angular*r_i*mean(r_i))
    total_var = var_x + var_y
    noise = np.sqrt(total_var / dof[:, None])
    # the same than np.std(yfft, axis=2) / np.sqrt(yfft.shape[2])
    residual_noise = noise[:, :, 0]
    # scale to match time domain amplitudes
    yfft = yfft / np.abs(yfft)
    angles = np.angle(np.sum(yfft * weights, axis=2) / np.sum(weights, axis=1))
    plv = np.abs(np.sum(yfft * weights, axis=2) / np.sum(weights, axis=1))
    z = dof * plv ** 2

    # D.Wilkie. Rayleigh Test for Randomness of Circular Data".Applied Statistics.1983.
    tmp = np.ones(z.shape)
    _idx_tmp = (dof < 50).squeeze()

    if np.any(_idx_tmp):
        _tmp_correction = 1.0 + (2.0 * z - z * z) / (4.0 * dof) - (
                24.0 * z - 132.0 * z ** 2.0 + 76.0 * z ** 3.0 - 9.0 * z ** 4.0) / (288.0 * dof * dof)
        tmp[:, _idx_tmp] = _tmp_correction
    z_crit = -np.log(alpha) - (2 * np.log(alpha) + np.log(alpha) ** 2.0) / (4 * dof)
    p_values = np.exp(-z) * tmp

    return spectral_amp, plv, z, z_crit, p_values, angles, dof, residual_noise



__author__ = 'jundurraga'

import numpy as np


def get_medel_am_alt_phase_biphasic_pulse_positions(stimulus_parameters={}, **kwargs):
    triggers = kwargs.get('triggers', np.array([]))
    trigger_to_first_pulse_delay = kwargs.get('trigger_to_first_pulse_delay', 8.333e-6)
    trigger_offset_correction = kwargs.get('trigger_offset_correction', True)
    fs = kwargs.get('fs', 16384.0)
    """

    :type trigger_positions: numpy array
    """
    _inter_pulse_interval = 1 / stimulus_parameters['CarrierRate']
    phase_alt_period = 1 / stimulus_parameters['AltPhaseRate']
    _comp_slope = 1.0
    _triggers_pos = triggers['idx']
    n_epochs = len(_triggers_pos)
    expected_samples_per_trigger = 0.0

    if trigger_offset_correction:
        expected_samples_per_trigger = stimulus_parameters['Duration']
        _trigger_distance = np.mean(np.diff(_triggers_pos)) / fs
        _comp_slope = _trigger_distance / expected_samples_per_trigger

    _n_pulses_per_epoch = np.round(stimulus_parameters['Duration'] * stimulus_parameters['CarrierRate'])
    _switch_count = 0
    _pulse_positions = np.array([])
    _temp_switch_distance = 0.0
    n_sequences = 1
    for i in np.arange(n_sequences):
        for j in np.arange(_n_pulses_per_epoch):
            _time = j * _inter_pulse_interval
            if j == 0:
                # set the starting phase for starting sequence
                _temp_switch_distance = _inter_pulse_interval * (stimulus_parameters['CarrierPhase'] / (2 * np.pi))
                if i > 0:
                    _temp_switch_distance = _inter_pulse_interval * (
                            stimulus_parameters['CarrierPhase'] - stimulus_parameters['CarrierAltPhase']) / (2 * np.pi)

            else:
                if np.round(_time * 1.0e6) >= np.round(phase_alt_period / 2 * (_switch_count + 1) * 1.0e6):
                    _switch = True
                    _switch_count += 1
                else:
                    _switch = False

                if _switch:
                    if np.mod(_switch_count, 2) == 1:
                        _switch_phase = _inter_pulse_interval * (stimulus_parameters['CarrierAltPhase'] -
                                                                 stimulus_parameters['CarrierPhase']) / (2 * np.pi)
                    else:
                        _switch_phase = _inter_pulse_interval * (stimulus_parameters['CarrierPhase'] -
                                                                 stimulus_parameters['CarrierAltPhase']) / (2 * np.pi)

                    _temp_switch_distance += _switch_phase

            _pulse_positions = np.append(_pulse_positions, j * _inter_pulse_interval + _temp_switch_distance)
            # compensate drift
            _pulse_positions[-1] = _comp_slope * (_pulse_positions[-1] - expected_samples_per_trigger / fs)

        # start sequence
        _sequence = np.ceil((trigger_to_first_pulse_delay + _pulse_positions) * fs)
        pulse_positions = _triggers_pos[0] + _sequence
    for i in np.arange(1, stimulus_parameters['NRepetitions'] * n_epochs):
        if i > len(_triggers_pos):
            break
        pulse_positions = np.append(pulse_positions, _sequence + _triggers_pos[i])
    assert isinstance(pulse_positions, np.ndarray)
    return pulse_positions


def get_medel_am_biphasic_pulse_positions(stimulus_parameters={}, **kwargs):
    triggers = kwargs.get('triggers', np.array([]))
    trigger_to_first_pulse_delay = kwargs.get('trigger_to_first_pulse_delay', 8.333e-6)
    trigger_offset_correction = kwargs.get('trigger_offset_correction', True)
    fs = kwargs.get('fs', 16384.0)
    """

    :type trigger_positions: numpy array
    """
    _inter_pulse_interval = 1 / stimulus_parameters['CarrierRate']
    _comp_slope = 1.0
    _triggers_pos = triggers['idx']
    n_epochs = len(_triggers_pos)
    expected_samples_per_trigger = 0.0

    if trigger_offset_correction:
        expected_samples_per_trigger = stimulus_parameters['Duration']
        _trigger_distance = np.mean(np.diff(_triggers_pos)) / fs
        _comp_slope = _trigger_distance / expected_samples_per_trigger

    _n_pulses_per_epoch = np.round(stimulus_parameters['Duration'] * stimulus_parameters['CarrierRate'])
    _pulse_positions = np.array([])
    _temp_switch_distance = 0.0
    for j in np.arange(_n_pulses_per_epoch):
        if j == 0:
            # set the starting phase for starting sequence
            _temp_switch_distance = _inter_pulse_interval * (stimulus_parameters['CarrierPhase'] / (2 * np.pi))
        _pulse_positions = np.append(_pulse_positions, j * _inter_pulse_interval + _temp_switch_distance)
        # compensate drift
        _pulse_positions[-1] = _comp_slope * (_pulse_positions[-1] - expected_samples_per_trigger / fs)

    _sequence = np.ceil((trigger_to_first_pulse_delay + _pulse_positions) * fs)
    pulse_positions = _triggers_pos[0] + _sequence
    for i in np.arange(1, stimulus_parameters['NRepetitions'] * n_epochs):
        if i > len(_triggers_pos):
            break
        pulse_positions = np.append(pulse_positions, _sequence + _triggers_pos[i])
    assert isinstance(pulse_positions, np.ndarray)
    return pulse_positions



# -*- coding: utf-8 -*-
"""
This module defines abstracts classes for general purpose EEG processing
@author:Jaime Undurraga

"""
import abc
from prettytable import PrettyTable
import logging
from peegy.processing.tools.epochs_processing_tools import et_snr_weighted_mean, et_get_spatial_filtering, \
    et_apply_spatial_filtering, get_in_window_data
import os
from os import listdir
from os.path import isfile, join
import json
import datetime
from scipy.io import savemat, loadmat
import io
import pickle
from peegy.processing.tools import epochs_processing_tools as et
from peegy.processing.tools.eeg_epoch_operators import et_fold, et_unfold
from peegy.processing.tools.detection.time_domain_tools import detect_peaks_and_amplitudes
from peegy.processing.tools.filters.eegFiltering import eeg_notch_filter, filt_filt_multithread
from peegy.definitions.eeg_definitions import EegAverageEpochs
from peegy.processing.tools.detection.definitions import PeakToPeakMeasure, TimePeakWindow, EegPeak, TimeROI
from peegy.definitions.eeg_definitions import EegChannel, ElectrodeType
from peegy.definitions.channel_definitions import Domain
from peegy.processing.pipe.spatial_filtering import SpatialFilter
from peegy.directories.tools import DirectoryPaths
from peegy.plot import eeg_ave_epochs_plot_tools as eegpt
from peegy.processing.statistics.eeg_statistic_tools import hotelling_t_square_test
import pandas
import multiprocessing
import ctypes
import scipy.signal as signal
import matplotlib.pyplot as plt
from sklearn import linear_model
import astropy.units as u
import numpy as np
log = logging.getLogger()


class EEGData(object, metaclass=abc.ABCMeta):
    def __init__(self, file_name='', **kwargs):
        delete_all = kwargs.get('delete_all', False)
        delete_figures = kwargs.get('delete_figures', False)
        self._file_name = file_name
        self.data_processing = {}
        self.data_raw = np.array([])
        self.epochs_raw = np.array([])
        self.epochs_raw_ave = EegAverageEpochs()
        self.epochs_processed = np.array([])
        self.epochs_processed_ave = EegAverageEpochs()
        self.component_maps_ave = [EegAverageEpochs()]
        self.ref_channels = np.array([EegChannel()])
        self.roi_windows = np.array([TimeROI()])
        self._reference = np.array([])
        self.spatial_filter = SpatialFilter()
        self.invert_polarity = False
        self.resampling_factor = 1.0
        self.low_pass_freq = None
        self.high_pass_freq = None
        self.notch_frequencies = None
        self.average_domain = Domain.time
        self._fs = None
        self._fs_raw = None
        self._paths = DirectoryPaths(file_path=file_name, delete_all=delete_all, delete_figures=delete_figures)
        self._oeg_channels = np.array([EegChannel()])
        self._bad_channels = np.array([EegChannel()])
        self.x_offset = 0.0
        self.n_epochs_to_read = None
        self.ini_epoch_to_read = None

    def get_file_name(self):
        return self._file_name

    def set_file_name(self, new_value):
        self._file_name = new_value

    file_name = property(get_file_name, set_file_name)

    def get_data(self):
        return 'set_data not defined'

    def set_data(self, new_value):
        return 'get_data not defined'

    data = abc.abstractproperty(get_data, set_data)

    def get_channel_labels(self):
        return 'get_channelLabels not defined'

    def set_channel_labels(self, new_value):
        return 'set_channelLabels not defined'

    channelLabels = abc.abstractproperty(get_channel_labels, set_channel_labels)

    def get_channels(self):
        return 'get_channels not defined'

    def set_channels(self):
        return 'set_channels not defined'

    channels = abc.abstractproperty(get_channels)

    def get_all_channels(self):
        return 'get_channels not defined'

    all_channels = abc.abstractproperty(get_all_channels)

    def get_fs(self):
        return self._fs

    def set_fs(self, new_value):
        self._fs = new_value

    fs = property(get_fs, set_fs)

    def get_oeg_channels(self):
        return self._oeg_channels

    def set_oeg_channels(self, new_value=np.array([EegChannel()])):
        self._oeg_channels = new_value

    oeg_channels = property(get_oeg_channels, set_oeg_channels)

    def get_bad_channels(self):
        return self._oeg_channels

    def set_bad_channels(self, new_value=np.array([EegChannel()])):
        self._bad_channels += new_value

    bad_channels = property(get_bad_channels, set_bad_channels)

    def get_amplitude_units(self):
        return 'get_fs not defined'

    def set_amplitude_units(self, new_value):
        return 'set_fs not defined'

    amplitude_units = abc.abstractproperty(get_amplitude_units, set_amplitude_units)

    def get_time_units(self):
        return 'get_fs not defined'

    def set_time_units(self, new_value):
        return 'set_fs not defined'

    time_units = abc.abstractproperty(get_time_units, set_time_units)

    def get_paths(self):
        return 'get_file_paths not defined'

    paths = abc.abstractproperty(get_paths)

    def get_system(self):
        return 'get_system not defined'

    system = abc.abstractproperty(get_system)

    def time_to_samples(self, value):
        out = None
        if value is not None:
            out = np.round(value * self.fs).astype(int)
        return out

    @abc.abstractmethod
    def get_raw_data(self, win_length=0):
        print("get_raw_data not defined")
        return "get_raw_data not defined"

    @abc.abstractmethod
    def get_data(self, channels=[]):
        print("getData not defined")
        return "getData not defined"

    @abc.abstractmethod
    def get_triggers(self, triggerCodes=[]):
        print('getTriggers not defined')
        return

    @abc.abstractmethod
    def get_channel_by_label(self, channel_labels=[]):
        print('get_channel_by_label not defined')
        return

    def set_reference(self):
        # if auto_detect_reference:
        #     ref_channels = [channels[self.detect_ref_channel(epochs)]]
        print(('Referencing data to: %s' % ' '.join(map(str, [_ch.label for _ch in self.ref_channels]))))
        reference = np.zeros(self.epochs_raw.shape[0::2], dtype=float)
        count = 0.0
        for ref_ch in self.ref_channels:
            for i, ch in enumerate(self.channels):
                if ch.__dict__ == ref_ch.__dict__:
                    reference += self.epochs_raw[:, i, :]
                    count += 1.0
                    #  remove reference if only one is provided
                    if len(self.ref_channels) == 1:
                        self.remove_channels(channels_to_remove=[ch])
                        print('Channel %s was removed but used as reference' % ch.label)
                    break
        reference /= np.maximum(count, 1.0)
        self._reference = reference
        t = PrettyTable()
        t.add_column(fieldname='Reference Channel', column=[_ref.label for _ref in self.ref_channels])
        log.info(t)
        print(t)

    def reference_epochs(self, epochs=np.array):
        # if auto_detect_reference:
        #     ref_channels = [channels[self.detect_ref_channel(epochs)]]
        print(('Referencing data to: %s' % ' '.join(map(str, [_ch.label for _ch in self.ref_channels]))))
        epochs -= np.expand_dims(self._reference, 1)
        return epochs

    def find_array(self, parameters={}, descriptor=''):
        parameters_files = [join(self.paths.data_subset_path, f) for f in
                            listdir(self.paths.data_subset_path)
                            if isfile(join(self.paths.data_subset_path, f)) and f.endswith('.json')]
        data = {}
        path = None
        for _file in parameters_files:
            with io.open(_file) as f:
                _configuration = json.loads(f.read())
                _file_par = _configuration['parameters']
                _parameters = dict(parameters, **{'descriptor': descriptor})
            if np.alltrue(
                    [_parameters[_key] == _file_par[_key] for _key in list(_parameters.keys()) if _key in _file_par]):
                path = _configuration['file']
                break
        if path:
            if not os.path.isfile(path):
                print(('could not find associated file: ' + path))
            else:
                print(('reading data from %s: ' % path))
                if os.path.splitext(path)[1] == '.hdf':
                    my_data = pandas.HDFStore(path)
                    for index, row in my_data['data_shape'].iterrows():
                        if (row['shape'] is not None) and len(row['shape']) == 3:
                            _data = et_fold(my_data[row['label']].as_matrix(), row['shape'][0])
                        else:
                            _data = my_data[row['label']].as_matrix()
                        data[row['label']] = _data
                    my_data.close()
                if os.path.splitext(path)[1] == '.mat':
                    data = loadmat(path)
        return data

    def save_array(self, parameters={}, data_type='array', data={}, data_format='.hdf', descriptor=''):
        if data_format not in ['.hdf', '.mat']:
            print(('unknown output format:' + data_format))
            return
        _c_date = datetime.datetime.today().strftime('%d_%b_%Y_%H_%M_%S')
        _parameters_file = self.paths.data_basename_path + descriptor + _c_date + '.json'
        _data_file = self.paths.data_basename_path + descriptor + _c_date
        out_dict = {'file': _data_file + data_format,
                    'parameters': dict(parameters, **{'descriptor': descriptor})}
        print(('saving data to %s' % _data_file + data_format))
        with io.open(_parameters_file, 'w', encoding='utf-8') as f:
            f.write(str(json.dumps(out_dict, ensure_ascii=False)))
        if data_format == '.hdf':
            x = pandas.HDFStore(_data_file + '.hdf', mode='w')
            x['parameters'] = pandas.DataFrame({'parameters': parameters})
            x['data_type'] = pandas.DataFrame({'data_type': [data_type]})
            data_shape = pandas.DataFrame(columns=('label', 'shape'))
            for _idx, _label in enumerate(data.keys()):
                if isinstance(data[_label], np.ndarray):
                    data_shape.loc[_idx] = [_label, data[_label].shape]
                    if len(data[_label].shape) == 3:
                        x[_label] = pandas.DataFrame(et_unfold(data[_label]))
                    else:
                        x[_label] = pandas.DataFrame(data[_label])
                else:
                    data_shape.loc[_idx] = [_label, None]
                    x[_label] = pandas.DataFrame(data[_label])
            x['data_shape'] = data_shape
            x.close()
            # np.save(_data_file, data)
        if data_format == '.mat':
            savemat(_data_file, data=data)

    def save_object(self, parameters={}, new_object=object, descriptor=''):
        _c_date = datetime.datetime.today().strftime('%d_%b_%Y_%H_%M_%S')
        _parameters_file = self.paths.data_basename_path + descriptor + _c_date + '.json'
        _data_file = self.paths.data_basename_path + descriptor + _c_date + '.p'
        out_dict = {'file': _data_file,
                    'parameters': parameters}
        with io.open(_parameters_file, 'w', encoding='utf-8') as f:
            f.write(str(json.dumps(out_dict, ensure_ascii=False)))
        with io.open(_data_file, 'wb') as f:
            pickle.dump(new_object, f)

    def find_object(self, parameters={}):
        parameters_files = [join(self.paths.data_subset_path, f)
                            for f in listdir(self.paths.data_subset_path)
                            if isfile(join(self.paths.data_subset_path, f)) and f.endswith('.json')]
        data = None
        path = None
        for _file in parameters_files:
            with io.open(_file) as f:
                _configuration = json.loads(f.read())
            if parameters == _configuration['parameters']:
                path = _configuration['file']
                break
        if path:
            if not os.path.isfile(path):
                print(('could not find associated file: ' + path))
            print(('reading object from: ' + path))
            with open(path, 'rb') as f:
                data = pickle.load(f)
        return data

    def set_data_raw(self,
                     ref_channels=np.array([EegChannel()]),
                     triggers=np.array([]),
                     **kwargs):
        pad_zeros = kwargs.get('pad_zeros', None)
        self.data_processing['pad_zeros'] = pad_zeros

        raw_data, relative_triggers, buffer_size = self.get_raw_data(channels=self.channels,
                                                                     ref_channels=ref_channels,
                                                                     triggers=triggers,
                                                                     **kwargs)
        # pad zeros if requested
        if pad_zeros is not None:
            int_zero_points = {'ini': relative_triggers - 1,
                               'end': relative_triggers + int(pad_zeros * self.fs)}
            self.interpolate_data(data=raw_data, interpolation_data_points=int_zero_points)
        self.data_raw = raw_data

    def get_epochs(self, channels=np.array([EegChannel()]), triggers={}, low_pass_freq=None,
                   high_pass_freq=None, notch_frequencies=None, **kwargs):
        epoch_length = kwargs.get('epoch_length', None)
        pad_zeros = kwargs.get('pad_zeros', None)
        resampling_fs = kwargs.get('resampling_fs', self._fs_raw)
        bad_channels = kwargs.get('bad_channels', [])
        remove_bad_channels = kwargs.get('remove_bad_channels', True)
        thr = kwargs.get('thr', None)
        interval = kwargs.get('interval', None)
        amp_thr = kwargs.get('amp_thr', None)
        descriptor = kwargs.get('descriptor', 'raw_epochs')
        trigger_factor = kwargs.get('trigger_factor', None)  # if provided, will add or remove triggers using this value

        # creates dictionary containing processing information
        self.data_processing['notch_frequencies'] = notch_frequencies
        self.data_processing['epoch_length'] = epoch_length
        self.data_processing['high_pass_freq'] = high_pass_freq
        self.data_processing['low_pass_freq'] = low_pass_freq
        self.data_processing['pad_zeros'] = pad_zeros
        self.data_processing['subset_identifier'] = self.paths.subset_identifier
        self.data_processing['ref_channels'] = [_ch.label for _ch in self.ref_channels]
        self.data_processing['resampling_fs'] = resampling_fs
        # self.data_processing['bad_channels'] = np.array([_bc.__dict__ for _bc in bad_channels])
        self.data_processing['remove_bad_channels'] = remove_bad_channels
        self.data_processing['thr'] = thr
        self.data_processing['interval'] = interval
        self.data_processing['amp_thr'] = amp_thr
        self.data_processing['x_offset'] = self.x_offset
        self.data_processing['trigger_factor'] = trigger_factor
        if descriptor == 'oeg_epochs':
            self.data_processing['oeg_channels'] = [_ch.label for _ch in self.oeg_channels]

        # creates dictionary containing processing information

        _data = self.find_array(parameters=self.data_processing, descriptor=descriptor)
        if _data:
            epochs = _data['epochs']
            _d_bad_channels = np.array([EegChannel(**b_c[0]) for b_c in _data['auto_detected_bad_channels']])
            self._fs = resampling_fs if resampling_fs else self._fs_raw
        else:
            data_raw, relative_triggers, buffer_size = self.get_raw_data(channels=channels,
                                                                         triggers=triggers,
                                                                         ref_channels=self.ref_channels,
                                                                         **kwargs)
            # correct triggers to compensate for x_offset
            relative_triggers += np.round(self.x_offset * self.fs).astype(int)

            # pad zeros if requested
            if pad_zeros is not None:
                int_zero_points = {'ini': relative_triggers - 1,
                                   'end': relative_triggers + int(pad_zeros * self.fs)}
                self.interpolate_data(data=data_raw, interpolation_data_points=int_zero_points)
            # detect bad channels
            _d_bad_channels = self.detect_bad_channels(epochs=data_raw, channels=channels, **kwargs)
            # filter data
            data = self.filter_data(data=data_raw, low_pass=low_pass_freq, high_pass=high_pass_freq)
            if self.notch_frequencies:
                data = eeg_notch_filter(data, f=notch_frequencies, f_range=1.0, fs=self._fs)

            # interpolate triggers
            if trigger_factor:
                relative_triggers = self.interpolate_triggers2(triggers=relative_triggers, factor=trigger_factor)
                buffer_size = int(np.maximum(buffer_size, np.min(np.diff(relative_triggers))))
                _useful = relative_triggers + buffer_size <= data.shape[0]
                relative_triggers = relative_triggers[_useful]
                t = PrettyTable()
                t.add_column(fieldname='Adjusted epoch length [s]', column=[buffer_size / self._fs])
                log.info(t)

            epochs = np.zeros((buffer_size, len(channels), len(relative_triggers)), dtype=np.float32)
            for i, trigger_pos in enumerate(relative_triggers):
                # ensure that blocks match buffer size
                if trigger_pos + buffer_size > data.shape[0]:
                    epochs = epochs[:, :, list(range(i))]
                    break
                epochs[:, :, i] = data[trigger_pos:trigger_pos + buffer_size, :]
            self.save_array(parameters=self.data_processing, data_type='array',
                            data={'epochs': epochs,
                                  'bad_channels': np.array([_bc.__dict__ for _bc in bad_channels]),
                                  'auto_detected_bad_channels': np.array([_bc.__dict__ for _bc in _d_bad_channels])},
                            descriptor=descriptor)
        t = PrettyTable()
        t.add_column(fieldname='File name', column=[self._file_name])
        t.add_column(fieldname='High-pass Filter', column=[str(high_pass_freq)])
        t.add_column(fieldname='Low-pass Filter', column=[str(low_pass_freq)])
        t.add_column(fieldname='Notch Filter', column=[str(notch_frequencies)])
        log.info(t)

        self.data_processing['ini_epoch_to_read'] = self.ini_epoch_to_read
        self.data_processing['n_epochs_to_read'] = self.n_epochs_to_read
        n_epochs = epochs.shape[2]
        ini_epoch = 0
        end_epoch = n_epochs
        if self.ini_epoch_to_read is not None:
            ini_epoch = np.minimum(self.ini_epoch_to_read, n_epochs)
        if self.n_epochs_to_read is not None:
            end_epoch = np.minimum(self.n_epochs_to_read + ini_epoch, n_epochs)
        epochs = epochs[:, :, ini_epoch:end_epoch]
        t = PrettyTable(['reading epochs'])
        t.add_row(['Reading {:d} epochs'.format(end_epoch - ini_epoch)])
        t.add_row(['from {:d}'.format(ini_epoch)])
        t.add_row(['to {:d}'.format(end_epoch - 1)])
        log.info(t)
        print(t)

        return epochs, np.concatenate((bad_channels, _d_bad_channels))

    def set_epochs_raw(self,
                       triggers={},
                       **kwargs):
        epoch_length = kwargs.pop('epoch_length', None)
        resampling_fs = kwargs.pop('resampling_fs', self._fs_raw)
        # get scalp data
        self.data_processing['channels'] = [_ch.label for _ch in self.channels]
        self.data_processing['layout_channel_number'] = [_ch for _ch in self.layout['channel']]
        self.data_processing['layout_channel_label'] = [_lb for _lb in self.layout['label']]
        self.epochs_raw, auto_detected_bad_channels = self.get_epochs(channels=self.channels,
                                                                      data_raw=self.data_raw,
                                                                      triggers=triggers,
                                                                      low_pass_freq=self.low_pass_freq,
                                                                      high_pass_freq=self.high_pass_freq,
                                                                      notch_filter=self.notch_frequencies,
                                                                      epoch_length=epoch_length,
                                                                      resampling_fs=resampling_fs,
                                                                      bad_channels=self.bad_channels,
                                                                      **kwargs)

        # detect bad channels across epochs
        _d_bad_channels = self.detect_bad_channels(epochs=self.epochs_raw, channels=self.channels, **kwargs)
        return [_ch for _ch in list(set([_o for _o in np.concatenate((auto_detected_bad_channels, _d_bad_channels))]))]

    def remove_oeg_artifacts(self, triggers={}, epoch_length=None, **kwargs):
        if self.oeg_channels.size:
            # separate oog channels from scalp channels
            # get oeg data
            oeg_epochs, _bad_channels = self.get_epochs(channels=self.oeg_channels,
                                                        epoch_length=epoch_length,
                                                        triggers=triggers,
                                                        high_pass_freq=0.3,
                                                        low_pass_freq=0.7,
                                                        resampling_fs=self._fs,
                                                        descriptor='oeg_epochs',
                                                        **kwargs)
            oeg_epochs = self.reference_epochs(epochs=oeg_epochs)
            # remove oeg artifacts
            epochs_raw = et.eye_artifact_subtraction(epochs=self.epochs_raw, oeg_epochs=oeg_epochs)
            self.epochs_raw = epochs_raw
            if 'oeg_channels' in list(self.data_processing.keys()):
                self.data_processing['oeg_channels'] += [_oeg.label for _oeg in self.oeg_channels]
            else:
                self.data_processing['oeg_channels'] = [_oeg.label for _oeg in self.oeg_channels]
            t = PrettyTable(['oeg channels subtracted'])
            [t.add_row([_oeg.label]) for _oeg in self.oeg_channels]
            log.info(t)
            print(t)

    def set_epochs_raw_ave(self):
        w_ave, w, rn, cum_rn, snr, cum_snr, s_var, w_fft, nk = \
            et.et_weighted_mean(epochs=self.epochs_raw,
                                block_size=max(np.floor(self.epochs_raw.shape[2] / 100), 10),
                                samples_distance=max(np.floor(self.epochs_raw.shape[0] / 256).astype(int), 10),
                                roi_windows=self.roi_windows_in_samples()
                                )

        raw_average = EegAverageEpochs(average=w_ave,
                                       channels=self.channels,
                                       rn=rn,
                                       cum_rn=cum_rn,
                                       snr=snr,
                                       cum_snr=cum_snr,
                                       n_samples_block=nk,
                                       signal_variance=s_var,
                                       fs=self.fs,
                                       amplitude_units=self.amplitude_units,
                                       time_units=self.time_units,
                                       invert_polarity=self.invert_polarity,
                                       data_processing=self.data_processing,
                                       roi_windows=self.roi_windows)
        self.epochs_raw_ave = raw_average

    def roi_windows_in_samples(self):
        roi_samples = []
        for _i, _roi in enumerate(self.get_unique_interval_rois()):
            roi_samples.append(self.time_to_samples(_roi.interval))
        return roi_samples

    def apply_spatial_filtering(self, sf_join_frequencies=None, sf_components=np.array([]), **kwargs):
        self.data_processing['sf_filtering'] = True
        _data = self.find_array(parameters=self.data_processing, descriptor='spatial_filter')
        if _data:
            z = _data['z']
            pwr0 = _data['pwr0']
            pwr1 = _data['pwr1']
            cov = _data['cov']
        else:
            # compute spatial filter
            z, pwr0, pwr1, cov = et_get_spatial_filtering(epochs=self.epochs_raw,
                                                          fs=self.fs,
                                                          sf_join_frequencies=sf_join_frequencies)
            # save the filter
            self.save_array(parameters=self.data_processing,
                            data_type='spatial_filter',
                            data={'z': z,
                                  'pwr0': pwr0,
                                  'pwr1': pwr1,
                                  'cov': cov},
                            descriptor='spatial_filter')
        # compute average components and attach them to spatial_filter
        z_ave, z_w, z_rn, z_cum_rn, z_snr, z_cum_snr, z_s_var, z_w_fft, nk = \
            et.et_weighted_mean(epochs=z,
                                block_size=max(np.floor(z.shape[2] / 100), 10),
                                samples_distance=max(np.floor(z.shape[0] / 256).astype(int), 10),
                                roi_windows=self.roi_windows_in_samples()
                                )

        components_ave = EegAverageEpochs(average=z_ave,
                                          channels=np.array(
                                              [EegChannel(number=_n) for _n in np.arange(z_ave.shape[1])]),
                                          rn=z_rn,
                                          cum_rn=z_cum_rn,
                                          snr=z_snr,
                                          cum_snr=z_cum_snr,
                                          n_samples_block=nk,
                                          signal_variance=z_s_var,
                                          fs=self.fs,
                                          invert_polarity=self.invert_polarity,
                                          amplitude_units=1,
                                          time_units=self.time_units,
                                          roi_windows=self.roi_windows)
        # apply spatial filter
        self.epochs_processed, sf_components = et_apply_spatial_filtering(z=z, pwr0=pwr0, pwr1=pwr1, cov_1=cov,
                                                                          sf_components=sf_components, **kwargs)
        self.spatial_filter.z = z
        self.spatial_filter.pwr0 = pwr0
        self.spatial_filter.pwr1 = pwr1
        self.spatial_filter.cov = cov
        self.spatial_filter.components_ave = components_ave
        self.spatial_filter.component_indexes = sf_components

        t = PrettyTable()
        t.add_column(fieldname='spatial filtering retained components',
                     column=sf_components)
        log.info(t)
        print(t)

    def set_epochs_processed_ave(self, test_frequencies=None):
        w_fft = np.array([])
        freq_tests = []
        f_peaks = []
        cum_rn = []
        cum_snr = []
        nk = None
        if self.average_domain == Domain.time:
            # average de-noised data across epochs
            w_ave, w, rn, cum_rn, snr, cum_snr, s_var, w_fft, nk = \
                et.et_weighted_mean(epochs=self.epochs_processed,
                                    block_size=max(np.floor(self.epochs_processed.shape[2] / 100), 10),
                                    samples_distance=max(
                                        np.floor(self.epochs_processed.shape[0] / 256).astype(int), 10),
                                    roi_windows=self.roi_windows_in_samples()
                                    )

        if self.average_domain == Domain.frequency:
            # average processed data across epochs including frequency average
            w_ave, rn, snr, w_fft, _ht_tests, s_var = \
                et.et_frequency_mean(epochs=self.epochs_processed,
                                     fs=self.fs,
                                     test_frequencies=test_frequencies,
                                     block_size=max(np.floor(self.epochs_processed.shape[2] / 100), 10),
                                     samples_distance=max(
                                         np.floor(self.epochs_processed.shape[0] / 256).astype(int), 10))

            # restructure frequency tests and channel label to them
            f_peaks, freq_tests = self.set_frequency_peaks(hotelling_tests=_ht_tests,
                                                           channels=self.channels,
                                                           s_var=s_var)
        # attach data to class
        self.epochs_processed_ave = EegAverageEpochs(average=w_ave,
                                                     rfft_average=np.abs(w_fft),
                                                     channels=self.channels,
                                                     rn=rn,
                                                     snr=snr,
                                                     cum_rn=cum_rn,
                                                     cum_snr=cum_snr,
                                                     n_samples_block=nk,
                                                     signal_variance=s_var,
                                                     fs=self.fs,
                                                     frequency_test=freq_tests,
                                                     peak_frequency=f_peaks,
                                                     invert_polarity=self.invert_polarity,
                                                     amplitude_units=self.amplitude_units,
                                                     time_units=self.time_units,
                                                     roi_windows=self.roi_windows)

    @staticmethod
    def set_frequency_peaks(hotelling_tests=None, channels=None, s_var=None):
        f_peaks = []
        freq_tests = []
        for _i, (_ht_ch, _ch) in enumerate(zip(hotelling_tests, channels)):
            for _s_ht in _ht_ch:
                _s_ht.label = _ch.label
                freq_tests.append(_s_ht)
                _f_peak = {'channel': _ch.label,
                           'x': _s_ht.frequency_tested,
                           'rn': _s_ht.rn,
                           'snr': _s_ht.snr,
                           'amp': _s_ht.spectral_magnitude,
                           'amp_snr': _s_ht.snr,
                           'significant': bool(_s_ht.p_value < 0.05),
                           'label_peak': "{:10.1f}".format(_s_ht.frequency_tested),
                           'show_label': True,
                           'positive': True,
                           'domain': Domain.frequency,
                           'spectral_phase': _s_ht.spectral_phase}

                f_peaks.append(EegPeak(**_f_peak))
        return f_peaks, freq_tests

    def get_unique_interval_rois(self):
        out = []
        if self.roi_windows is not None and len(self.roi_windows):
            intervals = np.array([_roi.interval for _roi in self.roi_windows])
            b = np.ascontiguousarray(intervals).view(np.dtype((np.void, intervals.dtype.itemsize * intervals.shape[1])))
            _, idx = np.unique(b, return_index=True)
            out = self.roi_windows[idx]
        return out

    def hotelling_t2_test(self, block_time=0.040, **kwargs):
        ht2_de_trend_data = kwargs.get('ht2_de_trend_data', True)
        block_size = int(block_time * self.fs)
        tests = np.array([])
        for _roi_idx, _roi in enumerate(self.get_unique_interval_rois()):
            _samples = self.time_to_samples(_roi.interval)
            data = self.epochs_processed.copy()
            # remove linear trend
            _ini = _samples[0]
            _end = _samples[1]
            if ht2_de_trend_data:
                x = np.expand_dims(np.arange(0, data[_ini:_end, :, :].shape[0]), axis=1)
                for _idx in np.arange(data.shape[2]):
                    _ini_dt = np.maximum(0, _ini - int(self.fs*0.1))
                    _end_dt = np.minimum(data.shape[0], _end + int(self.fs * 0.1))
                    _subset = data[_ini_dt:_end_dt, :, _idx].copy()
                    x_dt = np.expand_dims(np.arange(0, data[_ini_dt:_end_dt, :, :].shape[0]), axis=1)
                    regression = linear_model.LinearRegression()
                    regression.fit(x_dt, _subset)
                    data[_ini:_end, :, _idx] -= regression.predict(x)
            # remove mean
            data[_ini:_end, :, :] -= np.mean(data[_ini:_end, :, :], axis=0)
            block_size = max(0, min(block_size, _samples[1] - _samples[0]))
            n_blocks = np.floor((_samples[1] - _samples[0]) / block_size).astype(int)
            samples = np.array([np.mean(data[_ini + _i * block_size: _ini + (_i + 1) * block_size, :, :], 0)
                                for _i in range(n_blocks)])
            if self.epochs_processed_ave.cum_rn is not None:
                epoch_block_size = np.cumsum(self.epochs_processed_ave.n_samples_block)
                for _n, _rn in zip(epoch_block_size, self.epochs_processed_ave.cum_rn):
                    tests = np.append(tests, hotelling_t_square_test(samples[:, :, 0:_n],
                                                                     interval=_roi.interval,
                                                                     channels=self.channelLabels
                                                                     )
                                      )

            else:
                tests = np.append(tests, hotelling_t_square_test(samples))
        if tests is not []:
            for _test in tests:
                _test.detrend = ht2_de_trend_data
        return tests

    def set_individuals_components_maps(self, n_components=11, test_frequencies=None):
        components_maps = []
        n_components = min(n_components, self.spatial_filter.z.shape[1])
        for _component in np.arange(n_components):
            _filtered_data, _ = et_apply_spatial_filtering(z=self.spatial_filter.z, pwr0=self.spatial_filter.pwr0,
                                                           pwr1=self.spatial_filter.pwr1, cov_1=self.spatial_filter.cov,
                                                           sf_components=np.array([_component]))
            freq_tests = []
            w_fft = np.array([])
            f_peaks = []
            cum_rn = None
            cum_snr = None
            nk = None
            if self.average_domain == Domain.time:
                # average de-noised data across epochs
                w_ave, w, rn, cum_rn, snr, cum_snr, s_var, w_fft, nk = \
                    et.et_weighted_mean(epochs=_filtered_data,
                                        block_size=max(np.floor(_filtered_data.shape[2] / 100), 10),
                                        samples_distance=max(np.floor(_filtered_data.shape[0] / 256).astype(int), 10),
                                        roi_windows=self.roi_windows_in_samples()
                                        )
            if self.average_domain == Domain.frequency:
                # average processed data across epochs including frequency average
                w_ave, rn, snr, w_fft, _ht_tests, s_var = \
                    et.et_frequency_mean(epochs=_filtered_data,
                                         fs=self.fs,
                                         test_frequencies=test_frequencies,
                                         block_size=max(np.floor(_filtered_data.shape[2] / 100), 10),
                                         samples_distance=max(np.floor(_filtered_data.shape[0] / 256).astype(int), 10))
                # restructure frequency tests and channel label to them
                f_peaks, freq_tests = self.set_frequency_peaks(hotelling_tests=_ht_tests,
                                                               channels=self.channels,
                                                               s_var=s_var)

            comp_ave = EegAverageEpochs(average=w_ave,
                                        rfft_average=np.abs(w_fft),
                                        channels=self.channels[0:w_ave.shape[1]],
                                        rn=rn,
                                        cum_rn=cum_rn,
                                        snr=snr,
                                        cum_snr=cum_snr,
                                        n_samples_block=nk,
                                        signal_variance=s_var,
                                        fs=self.fs,
                                        frequency_test=freq_tests,
                                        peak_frequency=f_peaks,
                                        invert_polarity=self.invert_polarity,
                                        amplitude_units=self.amplitude_units,
                                        time_units=self.time_units,
                                        descriptor='component_{:d}'.format(_component),
                                        roi_windows=self.roi_windows
                                        )
            components_maps.append(comp_ave)
        self.component_maps_ave = components_maps

    def plot_individuals_components(self,
                                    eeg_topographic_map_channels=np.array([]),
                                    peak_time_windows=[TimePeakWindow()],
                                    component_maps=[EegAverageEpochs],
                                    x_lim_time=None,
                                    y_lim_time=None,
                                    x_lim_freq=None,
                                    y_lim_freq=None,
                                    domain=Domain.time,
                                    **kwargs):
        amplitude_unit = kwargs.get('amplitude_unit', 'microvolt')
        time_unit = kwargs.get('time_unit', 'second')
        fig_format = kwargs.get('fig_format', 'second')
        fontsize = kwargs.get('fontsize', 8.0)

        for _component, comp_ave in enumerate(component_maps):

            self.set_peaks_and_amplitudes(averaged_epochs=comp_ave,
                                          peak_time_windows=peak_time_windows,
                                          amp_sigma_thr=0.0)

            max_chan = np.argmax(np.std(comp_ave.average, axis=0))
            cha_max_pow_label = self.channels[max_chan].label
            # plot time potential fields for average channel
            if domain == Domain.time:
                eegpt.plot_eeg_topographic_map(ave_data=comp_ave,
                                               eeg_topographic_map_channels=[cha_max_pow_label],
                                               figure_dir_path=self.paths.figure_basename_path,
                                               figure_basename='component_time_' + str(_component),
                                               time_unit=time_unit,
                                               amplitude_unit=amplitude_unit,
                                               domain=Domain.time,
                                               title='Single component time domain projection: ' + str(_component),
                                               x_lim=x_lim_time,
                                               y_lim=y_lim_time,
                                               fig_format=fig_format,
                                               fontsize=fontsize)

            # plot frequency potential fields for average channel
            if domain == Domain.frequency:
                eegpt.plot_eeg_topographic_map(ave_data=comp_ave,
                                               eeg_topographic_map_channels=[cha_max_pow_label],
                                               figure_dir_path=self.paths.figure_basename_path,
                                               figure_basename='component_freq_' + str(_component),
                                               time_unit=time_unit,
                                               amplitude_unit=amplitude_unit,
                                               domain=Domain.frequency,
                                               title='Single component frequency domain projection: ' + str(_component),
                                               x_lim=x_lim_freq,
                                               y_lim=y_lim_freq,
                                               fig_format=fig_format,
                                               fontsize=fontsize)

    def append_average_channel(self, averaged_epochs=EegAverageEpochs(), channel_labels=np.array([])):
        across_channels_ave_r, total_rn_r, total_snr_r, s_var = \
            eeg_across_channel_snr_weighted_mean(ave_epochs=averaged_epochs,
                                                 channel_labels=channel_labels,
                                                 roi_windows=self.roi_windows_in_samples()
                                                 )
        # get polarity to add average channel with proper polarity
        pol = (-1.0) ** averaged_epochs.invert_polarity
        # append across channels average to data as another eeg channel

        label = '_'.join(channel_labels) if channel_labels.size else 'ave'

        averaged_epochs.append_eeg_ave_channel(new_channel_data=across_channels_ave_r * pol,
                                               rn=total_rn_r,
                                               snr=total_snr_r,
                                               signal_variance=s_var,
                                               label=label,
                                               electrode_type=ElectrodeType.artificial)

    def append_gfp_channel(self, averaged_epochs=EegAverageEpochs()):
        gfp_data, gfp_var, snr = self.eeg_gfp()
        averaged_epochs.append_eeg_ave_channel(new_channel_data=gfp_data,
                                               rn=0,
                                               snr=snr,
                                               signal_variance=gfp_var,
                                               label='gfp',
                                               electrode_type=ElectrodeType.artificial,
                                               fixed_polarity=True,
                                               roi_windows=self.roi_windows)

    @staticmethod
    def set_peaks_and_amplitudes(averaged_epochs=EegAverageEpochs(),
                                 peak_time_windows=[TimePeakWindow()],
                                 peak_to_peak_measures=[PeakToPeakMeasure()],
                                 **kwargs):
        peaks, amplitudes, new_time_windows = detect_peaks_and_amplitudes(
            epochs_ave=averaged_epochs,
            time_peak_windows=peak_time_windows,
            eeg_peak_to_peak_measures=peak_to_peak_measures,
            **kwargs)
        averaged_epochs.peak_times = peaks
        averaged_epochs.peak_amplitudes = amplitudes
        averaged_epochs.peak_time_windows = new_time_windows
        return new_time_windows

    def interpolate_data(self, data=np.array, **kwargs):
        interpolation_data_points = kwargs.get('interpolation_data_points', {})
        save_plots = kwargs.get('save_plots', True)
        if interpolation_data_points:
            print('interpolating data')
            idx_delete = np.concatenate((np.where(interpolation_data_points['ini'] > data.shape[0])[0],
                                         np.where(interpolation_data_points['end'] > data.shape[0])[0]))

            if np.any(idx_delete):
                interpolation_data_points['end'] = np.delete(interpolation_data_points['end'], idx_delete)
                interpolation_data_points['ini'] = np.delete(interpolation_data_points['ini'], idx_delete)

            if save_plots:
                fig, ax = plt.subplots(1, 1)
                ax.plot(data[:, 0])
            for _ini, _end in zip(interpolation_data_points['ini'], interpolation_data_points['end']):
                new_x = np.linspace(_ini, _end, num=_end - _ini + 1).astype(int)
                new_x_ar = np.tile(np.atleast_2d(new_x).T, (1, data.shape[1]))
                data[new_x, :] = (new_x_ar - _ini) * (data[_end, :] - data[_ini, :]) / (_end - _ini) + data[_ini, :]
            if save_plots:
                ax.plot(data[:, 0])
                m_point = len(interpolation_data_points['ini']) // 2
                for _i in range(10):
                    _fig_path = self.paths.figure_basename_path + '_interpolation_{:d}.png'.format(_i)
                    _ini = max(0, m_point - 2 * 2**_i)
                    _end = min(len(interpolation_data_points['ini']), m_point + 2 * 2**_i)
                    ax.set_xlim(interpolation_data_points['ini'][_ini], interpolation_data_points['ini'][_end])
                    ax.autoscale_view()
                    fig.savefig(_fig_path)
                plt.close(fig)

    @staticmethod
    def detect_bad_channels(epochs=np.array([]), channels=np.array([EegChannel()]), **kwargs):
        thr_sd_bad_channels = kwargs.get('thr_sd_bad_channels', 2)
        interval = kwargs.get('interval', 10)
        amp_thr = kwargs.get('amp_thr', 50000)
        if len(epochs.shape) == 3:
            data = et_unfold(epochs)
        else:
            data = epochs
        sub_data = data[np.arange(0, data.shape[0], interval), :]
        a_max = np.max(np.abs(sub_data), axis=0)
        bad_channels = np.where(np.abs(a_max) > amp_thr)[0]
        t = PrettyTable()
        t.add_column(fieldname='automatic bad channels detection',
                     column=['%s exceeds amplitude threshold of %s' %
                             (channels[_idx].label, amp_thr) for _idx in bad_channels])
        log.info(t)
        print(t)
        _others_idx = np.array([idx for idx in np.arange(sub_data.shape[1]) if idx not in bad_channels])
        a_std = np.std(sub_data[:, _others_idx], axis=0)
        thr_ci = thr_sd_bad_channels * np.std(a_std) + np.mean(a_std)
        n_ch_idx = np.where(a_std > thr_ci)[0]
        bad_idx = _others_idx[n_ch_idx] if n_ch_idx.size else np.array([], dtype=int)
        t = PrettyTable()
        t.add_column(fieldname='automatic bad channels detection',
                     column=['%s considered an outlier, across channels std = %f, channel std (%f) > %f * std (%f)' %
                             (channels[_c_idx].label, np.std(a_std), a_std[_c_n_ch] - np.mean(a_std),
                              thr_sd_bad_channels,
                              thr_ci - np.mean(a_std))
                             for _c_idx, _c_n_ch in zip(bad_idx, n_ch_idx)])
        log.info(t)
        print(t)
        bad_channels = np.concatenate((bad_channels, bad_idx))
        _bad_channels_index = np.unique(bad_channels)
        bad_channel_list = [channels[_idx] for _idx in _bad_channels_index]
        return bad_channel_list

    def remove_channels(self, channels_to_remove=np.array([EegChannel()]), keep=np.array([EegChannel()])):
        if (isinstance(keep, list) and len(keep) == 0) or keep is None:
            keep = np.array([EegChannel()])

        channels_to_remove = [_b_ch for _b_ch in channels_to_remove for _k in keep if _k.__dict__ != _b_ch.__dict__]
        t = PrettyTable(['Removed Channels'])
        for _in_ch in channels_to_remove:
            for _idx, _ch in enumerate(self.channels):
                if _ch.label == _in_ch.label:
                    self.channels = np.delete(self.channels, _idx)
                    self.epochs_raw = np.delete(self.epochs_raw, _idx, 1)
                    t.add_row([_ch.label])
                    if 'removed_bad_channels' in list(self.data_processing.keys()):
                        self.data_processing['removed_bad_channels'] += [_ch.label]
                    else:
                        self.data_processing['removed_bad_channels'] = [_ch.label]
                    break
        log.info(t)
        print(t)

    @staticmethod
    def swap_channels(data=np.array([]),
                      channels=np.array([EegChannel()]),
                      channel_labels=[('label_from', 'label_to')]):
        for _left, _right in channel_labels:
            [(ch_left, pos_from)] = [(_ch, _pos) for _pos, _ch in enumerate(channels) if _ch.label == _left]
            [(ch_right, pos_to)] = [(_ch, _pos) for _pos, _ch in enumerate(channels) if _ch.label == _right]
            if not (ch_left and ch_right):
                print("could not find %s and %s pair of channel to perform swap" % (_left, _right))
                continue
            ch_left.label = _right
            ch_right.label = _left
            if data.ndim == 2:
                data[:, [pos_from, pos_to]] = data[:, [pos_to, pos_from]]
            if data.ndim == 3:
                data[:, [pos_from, pos_to], :] = data[:, [pos_to, pos_from], :]

    @staticmethod
    def remove_channels_from_channel_list(channels_to_remove=np.array([EegChannel()]),
                                          channels=np.array([EegChannel()])):
        _str_list = ' '.join(map(str, [_ch.label for _ch in channels]))
        for _in_ch in channels_to_remove:
            for _idx, _ch in enumerate(channels):
                if _ch.label == _in_ch.label:
                    channels.pop(_idx)
                    print('%s channel was removed from a list of channels: %s' % (_ch.label, _str_list))
                    break

    @staticmethod
    def detect_ref_channel(data, interval=10):
        # find best reference electrode
        rank = []
        for i in range(data.shape[1]):
            _var = []
            a = data[np.arange(0, data.shape[0], interval), :] - \
                np.expand_dims(data[np.arange(0, data.shape[0], interval), i], 1)
            for j in np.arange(a.shape[1]):
                _var.append(np.var(a[:, j], ddof=1, axis=0))
            rank.append(np.std(_var))
        return np.argmin(rank)

    def filter_data(self, data=np.array([]), blocks=None, low_pass=None, high_pass=None, plot_filter_response=False):
        if not low_pass and not high_pass:
            return data
        mp_arr = multiprocessing.Array(ctypes.c_double, data.size, lock=False)
        arr = np.frombuffer(mp_arr)
        filtered_signal = arr.reshape(data.shape)
        np.copyto(filtered_signal, data)

        _b_l = None
        _a_l = None

        if low_pass is not None:
            low = 2 * low_pass / self._fs
            _b_l, _a_l = signal.butter(N=3, Wn=low, btype='lowpass', analog=False)
            filt_filt_multithread(filtered_data=filtered_signal, data=data, b=_b_l, a=_a_l)
        if high_pass is not None:
            high = 2 * high_pass / self._fs
            _b_h, _a_h = signal.butter(N=3, Wn=high, btype='highpass', analog=False)
            filt_filt_multithread(filtered_data=filtered_signal, data=filtered_signal, b=_b_h, a=_a_h)

        if plot_filter_response and (low_pass or high_pass is not None):
            w, h = signal.freqz(_b_l, _a_l, 4096)
            plt.plot(self._fs / 2.0 * w / np.pi, 20 * np.log10(abs(h)))
            plt.xscale('log')
            plt.xlabel('Frequency [Hz]')
            plt.ylabel('Amplitude [dB]')
            plt.margins(0, 0.1)
            plt.grid(which='both', axis='both')
            plt.show()
            w, h = signal.freqz(_b_h, _a_h, 4096)
            plt.plot(self._fs / 2.0 * w / np.pi, 20 * np.log10(abs(h)))
            plt.xscale('log')
            plt.xlabel('Frequency [Hz]')
            plt.ylabel('Amplitude [dB]')
            plt.margins(0, 0.1)
            plt.grid(which='both', axis='both')
            plt.show()

            plt.plot(np.hstack((filtered_signal, data)))
            plt.show()
            plt.plot(np.arange(data.shape[0]) * self._fs / data.shape[0],
                     np.abs(np.fft.fft(np.hstack((data, filtered_signal)), axis=0)))
            plt.show()

        return filtered_signal

    @staticmethod
    def interpolate_triggers2(triggers=np.array([]), factor=2):
        idx = np.array([], dtype=int)
        if factor > 1:
            for _i, (_ini, _end) in enumerate(zip(triggers[0:-1], triggers[1:])):
                _samples = np.linspace(_ini, _ini + (_end - _ini) / factor, factor).astype(int)
                idx = np.append(idx, _samples)
            #  append last items
            mean_trigger_diff = np.round(np.mean(np.diff(triggers, axis=0)))
            _samples = np.linspace(triggers[-1], triggers[-1] + mean_trigger_diff / factor, factor).astype(int)
            idx = np.append(idx, _samples)
        else:
            for _i, _c_idx in enumerate(triggers[0:-1: int(1 / factor)]):
                idx = np.append(idx, _c_idx)
        t = PrettyTable()
        t.add_column(fieldname='Initial number of triggers', column=[len(triggers)])
        t.add_column(fieldname='Final number of triggers (interpolated)', column=[len(idx)])
        log.info(t)
        print(t)
        return idx

    @staticmethod
    def interpolate_triggers(triggers={}, factor=2):
        code = np.array([], dtype=int)
        idx = np.array([], dtype=int)
        dur = np.array([])
        dur_samples = np.array([], dtype=int)
        max_distance = np.array([], dtype=int)
        min_distance = np.array([], dtype=int)
        position = np.array([])
        if factor > 1:
            for _i, (_ini, _end) in enumerate(zip(triggers['idx'][0:-1], triggers['idx'][1:])):
                _samples = np.linspace(_ini, _ini + (_end - _ini) / factor, factor).astype(int)
                idx = np.append(idx, _samples)
                dur = np.append(dur, [triggers['idx'][_i]]*len(_samples))
                dur_samples = np.append(dur_samples, [triggers['dur_samples'][_i]]*len(_samples))
                code = np.append(code, [triggers['code'][_i]] * len(_samples))
                position = np.append(position, [triggers['position'][_i]] * len(_samples))
        else:
            for _i, _c_idx in enumerate(triggers['idx'][0:-1: int(1/factor)]):
                idx = np.append(idx, _c_idx)
                dur = np.append(dur, [triggers['idx'][_i]])
                dur_samples = np.append(dur_samples, [triggers['dur_samples'][_i]])
                code = np.append(code, [triggers['code'][_i]])
                position = np.append(position, [triggers['position'][_i]])

        max_distance = [np.max(np.diff(idx))] * len(idx)
        min_distance = [np.min(np.diff(idx))] * len(idx)
        out = {'idx': idx,
               'dur': dur,
               'dur_samples': dur_samples,
               'max_distance': max_distance,
               'min_distance': min_distance,
               'code': code,
               'position': position,
               'triggers': triggers['triggers']}
        return out

    def eeg_gfp(self):
        gfp = np.std(self.epochs_processed_ave.average, axis=1).reshape((-1, 1))
        roi_samples = self.roi_windows_in_samples()
        _data_rois = get_in_window_data(data=gfp, roi_windows=roi_samples)
        s_var = []
        snr = []
        for _data in _data_rois:
            s_var.append(np.var(_data, ddof=1, axis=0))
            snr.append(np.Inf * u.dimensionless_unscaled)
        return gfp, s_var, snr


def eeg_across_channel_snr_weighted_mean(ave_epochs=EegAverageEpochs(), channel_labels=[],
                                         roi_windows=np.array([])):
    channel_labels = np.array(channel_labels)
    is_scalp_ch = ave_epochs.is_scalp_channel()

    if channel_labels.size:
        _idx_channels = np.array([_i_ch for _label in channel_labels for _i_ch, _ch in
                                  enumerate(ave_epochs.channels) if _label == _ch.label and is_scalp_ch[_i_ch]])
    else:
        _idx_channels = np.array([_i_ch for _i_ch, _ch in enumerate(ave_epochs.channels) if is_scalp_ch[_i_ch]])

    across_channels_ave_r, total_rn_r, total_snr_r, s_var = \
        et_snr_weighted_mean(averaged_epochs=ave_epochs.average,
                             fs=ave_epochs.fs,
                             rn=ave_epochs.rn,
                             snr=ave_epochs.get_max_snr_per_channel(),
                             channel_idx=_idx_channels,
                             roi_windows=roi_windows)
    return across_channels_ave_r, total_rn_r, total_snr_r, s_var



import os
from os.path import sep
import shutil


class DirectoryPaths(object):
    def __init__(self,
                 file_path: str = '',
                 delete_all: bool = False,
                 delete_figures: bool = False,
                 figures_folder: str = None,
                 figures_subset_folder=''):
        file_path = os.path.expanduser(file_path)
        self.file_path = file_path
        self.file_directory = os.path.dirname(os.path.abspath(file_path))
        self.file_basename_path = os.path.abspath(os.path.splitext(file_path)[0])
        self.delete_all = delete_all
        self.delete_figures = delete_figures
        self.figures_folder = figures_folder
        if self.figures_folder is None:
            self.figures_folder = os.path.join(self.file_directory, 'figures')
        if (self.delete_all or self.delete_figures) and os.path.exists(self.figures_folder):
            try:
                shutil.rmtree(self.figures_folder)
            except OSError:
                print((OSError.message))
        if not os.path.exists(self.figures_folder):
            os.makedirs(self.figures_folder)

        data_directory = os.path.join(self.file_directory, '.data')
        if self.delete_all and os.path.exists(data_directory):
            try:
                shutil.rmtree(data_directory)
            except OSError:
                print((OSError.message))
        if not os.path.exists(data_directory):
            os.makedirs(data_directory)
        self.data_dir = data_directory
        self.figures_subset_folder = figures_subset_folder

    def get_figure_basename_path(self):
        return os.path.join(self.figure_subset_path, os.path.basename(self.file_path).split('.')[0])

    figure_basename_path = property(get_figure_basename_path)

    def get_data_basename_path(self):
        return os.path.join(self.data_subset_path, os.path.basename(self.file_path).split('.')[0])

    data_basename_path = property(get_data_basename_path)

    def get_figure_subset_path(self):
        _path = os.path.join(self.figures_folder, self.figures_subset_folder) + sep
        if not os.path.exists(_path):
            os.makedirs(_path)
        return _path

    figure_subset_path = property(get_figure_subset_path)

    def get_data_subset_path(self):
        _path = os.path.join(self.data_dir, self.figures_subset_folder)
        if not os.path.exists(_path):
            os.makedirs(_path)
        return _path

    data_path = property(get_data_subset_path)
    figures_current_dir = property(get_figure_subset_path)



"""
.. _tut-assr-ht2-test-sim-wrong-bias:

############################################################################################
ASSR Hotelling-T2 test with spatial filtering biased with wrong frequencies(Simulated)
############################################################################################

In this example we simulate an ASSR and we assess its significance using the Hotelling's T2 test.
However, we used DSS with a bias frequency which does not correspond to the ASSR frequency. This is a wrong bias and
the example illustrates how this may significantly reduce the amplitude of the ASSR.
This method uses the real and imaginary part of the frequency bin of interest to assess whether the points from
each epoch are significantly different from zero (the center of the polar complex-plane).

"""
# Enable below for interactive backend
# import matplotlib
# if 'Qt5Agg' in matplotlib.rcsetup.all_backends:
#    matplotlib.use('Qt5Agg')
from peegy.processing.pipe.pipeline import PipePool
from peegy.processing.pipe.definitions import Events, Domain
from peegy.processing.pipe.general import ReferenceData, FilterData, AutoRemoveBadChannels
from peegy.processing.pipe.epochs import AverageEpochsFrequencyDomain, EpochData
from peegy.processing.pipe.attach import AppendGFPChannel
from peegy.processing.pipe.statistics import FTest
from peegy.processing.pipe.plot import PlotWaveforms, PlotTopographicMap
from peegy.processing.pipe.spatial_filtering import CreateAndApplySpatialFilter
from peegy.processing.pipe.simulate import GenerateInputData
from peegy.processing.tools.template_generator.auditory_waveforms import aep
from peegy.processing.pipe.storage import MeasurementInformation, SubjectInformation, SaveToDatabase
import os
import astropy.units as u
import numpy as np


# %%
# Generate some data
# ===================
# We generate some auditory steady-state response (ASSR)

fs = 256.0 * u.Hz
epoch_length = 4.0 * u.s
epoch_length = np.ceil(epoch_length * fs) / fs  # fit to fs rate
assr_frequency = 41.0 * u.Hz
assr_frequency = np.ceil(epoch_length * assr_frequency) / epoch_length  # fit to epoch length
# we select several unrelated frequencies to the ASSR to bias the filter
wrong_bias = np.array([13, 26, 39]) * u.Hz
test_frequencies = np.concatenate((
    [assr_frequency,
     2 * assr_frequency],
    wrong_bias
))
template_waveform, _ = aep(fs=fs)
n_channels = 32
event_times = np.arange(0, 360.0, 1 / assr_frequency.to(u.Hz).value) * u.s
reader = GenerateInputData(template_waveform=template_waveform,
                           fs=fs,
                           n_channels=n_channels,
                           mixing_matrix=np.diag(np.arange(n_channels))/n_channels,
                           snr=0.05,
                           layout_file_name='biosemi32.lay',
                           event_times=event_times,
                           event_code=1.0,
                           noise_attenuation=0,
                           return_noise_only=False,
                           figures_subset_folder='assr_ht2_test_wrong_bias')
reader.run()
# %%
# Resize events
# =======================
# Now we keep events at intervals that correspond to our desired epoch length


events = reader.output_node.events.get_events(code=1)
# skip events to preserve only those at each epoch point
_new_events = Events(events=events[0:-1:int(epoch_length * assr_frequency)])
reader.output_node.events = _new_events


# %%
# Start the pipeline
# ============================
# Some processing to obtain clean epochs


pipeline = PipePool()
pipeline['referenced'] = ReferenceData(reader,
                                       reference_channels=['Cz'],
                                       invert_polarity=True)
pipeline['channel_cleaned'] = AutoRemoveBadChannels(pipeline['referenced'])
pipeline['time_filtered_data'] = FilterData(pipeline['channel_cleaned'],
                                            high_pass=0.05 * u.Hz,
                                            low_pass=None)
pipeline['time_epochs'] = EpochData(pipeline['time_filtered_data'],
                                    event_code=1.0,
                                    base_line_correction=False)
pipeline.run()

# %%
# Compute and plot components in the frequency-domain
# ----------------------------------------------------
# Spatial filter is a applied in the frequency-domain

pipeline['dss_time_epochs'] = CreateAndApplySpatialFilter(pipeline['time_epochs'],
                                                          sf_join_frequencies=wrong_bias,
                                                          test_frequencies=test_frequencies,
                                                          projection_domain=Domain.frequency,
                                                          plot_y_lim=[0, 1],
                                                          return_figures=True
                                                          )
pipeline.run()

# %%
# Compute global field power (GFP)
# ----------------------------------
# We compute the GFP across channels and epochs. A new channel with the GFP is appended to the data


pipeline['dss_time_epochs_with_gfp'] = AppendGFPChannel(pipeline['dss_time_epochs'])
pipeline.run()

# %%
# Average epochs
# ----------------------------------
# We compute the average and simultaneously get statistical tests on the test_frequencies


pipeline['fft_ave'] = AverageEpochsFrequencyDomain(pipeline['dss_time_epochs_with_gfp'],
                                                   n_fft=int(epoch_length*fs),
                                                   test_frequencies=test_frequencies)

pipeline['fft_ave_f_test'] = FTest(pipeline['fft_ave'],
                                   test_frequencies=test_frequencies,
                                   delta_frequency=15 * u.Hz)

pipeline['std_fft_ave'] = AverageEpochsFrequencyDomain(pipeline['time_epochs'],
                                                       n_fft=int(epoch_length * fs),
                                                       test_frequencies=test_frequencies)

pipeline.run()

pipeline['fft_ave_f_test'].output_node.statistical_tests['f_test_freq'][
    ["test_name", "df_1", "df_2", "f", "f_critic", "p_value"]].head()

# %%
# Generate figures
# ----------------------------------
# Now we run plot the average waveforms and show the stats


pipeline['waveforms'] = PlotWaveforms(pipeline['std_fft_ave'],
                                      overlay=[pipeline['fft_ave']],
                                      plot_x_lim=[0, 90],
                                      ch_to_plot=np.array(['O2', 'T8', 'T7']),
                                      user_naming_rule='standard_and_dss',
                                      statistical_test='hotelling_t2_freq',
                                      show_following_stats=['test_name', 'f', 'frequency_tested'],
                                      return_figures=True)

pipeline['topographic_map'] = PlotTopographicMap(pipeline['fft_ave'],
                                                 topographic_channels=np.array(['O2', 'T8', 'T7', 'GFP']),
                                                 plot_x_lim=[0, 90],
                                                 plot_y_lim=[0, 6],
                                                 return_figures=True)
pipeline.run()
# %%
# Save results to a database
# ----------------------------------
# We get the measurements we are interested in and save them into a database

subject_info = SubjectInformation(subject_id='Test_Subject')
measurement_info = MeasurementInformation(
    date='Today',
    experiment='sim')

_parameters = {'Type': 'ASSR'}
database_path = reader.input_node.paths.file_directory + os.sep + 'assr_ht2_test_wrong_bias.sqlite'
pipeline['database'] = SaveToDatabase(database_path=database_path,
                                      measurement_information=measurement_info,
                                      subject_information=subject_info,
                                      recording_information={'recording_device': 'dummy_device'},
                                      stimuli_information=_parameters,
                                      processes_list=[pipeline['fft_ave']],
                                      include_waveforms=True
                                      )
pipeline.run()

# %%
# Generate pipeline diagram
# ------------------------------------
pipeline.diagram(file_name=reader.output_node.paths.figures_current_dir + 'pipeline.png',
                 return_figure=True,
                 dpi=600)



from scipy.signal import hilbert, group_delay
import astropy.units as u
import numpy as np
from peegy.definitions.channel_definitions import Domain
from peegy.processing.pipe.definitions import InputOutputProcess
from peegy.processing.tools.detection.definitions import TimeROI
from scipy import signal


class InstantaneousPhase(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 roi_windows: np.array([TimeROI]) = None,
                 relative_difference: bool = True,
                 reference_frequency: u.Quantity = None,
                 reference_time: u.Quantity = None,
                 **kwargs):
        """
        This InputOutputProcess computes the instantaneous phase of the data in  the inputprocess
        :param input_process: InputOutputProcess Class
        :param roi_windows: time windows used to perform some measures (snr, rms)
        :param relative_difference: if True, the difference using the initial phase as a reference will be returned
        :param reference_frequency: frequency to compute the relative phase difference
        :param reference_time: time point used to extract reference instantaneous phase
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(InstantaneousPhase, self).__init__(input_process=input_process, **kwargs)
        self.roi_windows = roi_windows
        self.relative_difference = relative_difference
        self.reference_frequency = reference_frequency
        self.reference_time = np.array([reference_time.value]) * reference_time.unit

    def transform_data(self):
        analytic_signal = hilbert(self.input_node.data, axis=0)
        instantaneous_phase = np.unwrap(np.angle(analytic_signal), axis=0) * u.rad
        if self.relative_difference and self.reference_frequency is not None:
            ref_index = self.input_node.x_to_samples(self.reference_time)
            ref = np.cos(2 * np.pi * self.reference_frequency * u.rad * self.input_node.x[:, None] *
                         np.ones(self.input_node.data.shape) + instantaneous_phase[ref_index, :])
            relative_analytic_signal = hilbert(ref, axis=0)
            relative_instantaneous_phase = np.unwrap(np.angle(relative_analytic_signal), axis=0) * u.rad
            instantaneous_phase = relative_instantaneous_phase - instantaneous_phase
            instantaneous_phase = instantaneous_phase - instantaneous_phase[ref_index, :]
        self.output_node.data = instantaneous_phase
        # for the moment we use the rms of the entire epoch as a noise measurement
        self.output_node.rn = np.sqrt(np.mean(instantaneous_phase ** 2, axis=0))


class InstantaneousEnvelope(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 roi_windows: np.array([TimeROI]) = None,
                 **kwargs):
        """
        This InputOutputProcess computes the instantaneous phase of the data in  the inputprocess
        :param input_process: InputOutputProcess Class
        :param roi_windows: time windows used to perform some measures (snr, rms)
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(InstantaneousEnvelope, self).__init__(input_process=input_process, **kwargs)
        self.roi_windows = roi_windows

    def transform_data(self):
        analytic_signal = hilbert(self.input_node.data, axis=0)
        amplitude_envelope = np.abs(analytic_signal) * self.input_node.data.unit
        self.output_node.data = amplitude_envelope
        # for the moment we use the rms of the entire epoch as a noise measurement
        self.output_node.rn = np.sqrt(np.mean(amplitude_envelope ** 2, axis=0))


class InstantaneousFrequency(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 roi_windows: np.array([TimeROI]) = None,
                 **kwargs):
        """
        This InputOutputProcess computes the instantaneous frequency of the data in  the inputprocess
        :param input_process: InputOutputProcess Class
        :param roi_windows: time windows used to perform some measures (snr, rms)
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(InstantaneousFrequency, self).__init__(input_process=input_process, **kwargs)
        self.roi_windows = roi_windows

    def transform_data(self):
        analytic_signal = hilbert(self.input_node.data, axis=0)
        instantaneous_phase = np.unwrap(np.angle(analytic_signal), axis=0)
        instantaneous_frequency = (np.diff(instantaneous_phase, axis=0) /
                                   (2.0 * np.pi) * self.input_node.fs)
        self.output_node.data = instantaneous_frequency * u.dimensionless_unscaled


class GroupDelay(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 roi_windows: np.array([TimeROI]) = None,
                 frequencies: np.array(u.Quantity) = None,
                 **kwargs):
        """
        This InputOutputProcess computes the instantaneous frequency of the data in  the inputprocess
        :param input_process: InputOutputProcess Class
        :param roi_windows: time windows used to perform some measures (snr, rms)
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(GroupDelay, self).__init__(input_process=input_process, **kwargs)
        self.roi_windows = roi_windows
        self.frequencies = frequencies

    def transform_data(self):
        if self.frequencies is None:
            self.frequencies = np.fft.rfftfreq(self.input_node.data.shape[0], 1 / self.input_node.fs)
            n_fft = self.input_node.data.shape[0]
        else:
            n_fft = self.frequencies.shape[0]
        gd = np.zeros((self.frequencies.shape[0], self.input_node.data.shape[1]))
        for _ch in range(self.input_node.data.shape[1]):
            w, g = group_delay((self.input_node.data[:, _ch], 1), w=self.frequencies.value,
                               fs=self.input_node.fs.value)
            gd[:, _ch] = g
        self.output_node.n_fft = n_fft
        self.output_node.data = gd / self.input_node.fs
        self.output_node.domain = Domain.frequency


class RelativeInstantaneousPhaseFrequency(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 reference_roi_window: TimeROI = None,
                 relative_difference: bool = True,
                 reference_frequency: u.Quantity = None,
                 wavelet_half_amplitude_time: u.Quantity = 80 * u.ms,
                 align_reference_phase: bool = True,
                 align_reference_time: u.Quantity = 0 * u.s,
                 **kwargs):
        """
        This InputOutputProcess computes the instantaneous phase of the data in  the inputprocess
        :param input_process: InputOutputProcess Class
        :param reference_roi_windows: reference time window to center phase
        :param relative_difference: if True, the difference using the initial phase as a reference will be returned
        :param reference_frequency: frequency to compute the relative phase difference
        :param wavelet_half_amplitude_time: full width half maximum length (in time) of the wavelet half-amplitude
        :param align_reference_phase: if true, the starting phase will by aligned to a given time given by
        align_reference_time
        :param align_reference_time: reference time used to align the phase of reference signal to the the input signal
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(RelativeInstantaneousPhaseFrequency, self).__init__(input_process=input_process, **kwargs)
        self.reference_roi_window = reference_roi_window
        self.relative_difference = relative_difference
        self.reference_frequency = reference_frequency
        self.wavelet_half_amplitude_time = wavelet_half_amplitude_time
        self.align_reference_phase = align_reference_phase
        self.align_reference_time = align_reference_time

    def transform_data(self):
        s = (self.wavelet_half_amplitude_time.to(u.s) / 2 * self.input_node.fs.to(u.Hz))
        w = self.reference_frequency * (2 * s * np.pi) / self.input_node.fs
        y_filter = np.zeros(self.input_node.data.shape, dtype=complex)
        reference_signal = (np.cos(2 * np.pi * self.reference_frequency * u.rad * self.input_node.x) +
                            1j * np.sin(2 * np.pi * self.reference_frequency * u.rad * self.input_node.x))[:, None]
        reference_signal = reference_signal / np.abs(reference_signal)
        if self.input_node.data.ndim == 1:
            y_filter = signal.cwt(self.input_node.data.value, signal.morlet2, widths=[s], w=w)
        if self.input_node.data.ndim == 2:
            for _i in range(y_filter.shape[1]):
                y_filter[:, _i] = signal.cwt(self.input_node.data[:, _i].value, signal.morlet2,
                                             widths=[s], w=w)
        if self.input_node.data.ndim == 3:
            for _i in range(y_filter.shape[1]):
                for _j in range(y_filter.shape[2]):
                    y_filter[:, _i, _j] = signal.cwt(self.input_node.data[:, _i, _j].value, signal.morlet2,
                                                     widths=[s], w=w)
        normalized_vector = y_filter / (np.abs(y_filter)) * u.dimensionless_unscaled

        if self.align_reference_phase:
            _reference_signal = reference_signal * np.ones(normalized_vector.shape)
            _ref_sample = self.input_node.x_to_samples([self.align_reference_time])
            _phase = np.angle(normalized_vector[_ref_sample[0], ...])
            _ref_phase = np.angle(_reference_signal[_ref_sample[0], ...])
            _delta_phase = _phase - _ref_phase
            reference_signal = _reference_signal * np.exp(1j * _delta_phase.value)
        if self.input_node.data.ndim == 3:
            normalized_vector = np.mean(normalized_vector, axis=2)
        # get each phase
        reference_phase = np.unwrap(np.angle(reference_signal), axis=0)
        signal_phase = np.unwrap(np.angle(normalized_vector), axis=0)
        instantaneous_phase_difference = reference_phase - signal_phase
        if self.reference_roi_window is not None:
            _samples = self.reference_roi_window.get_roi_samples(self.input_node)
            reference_samples = np.arange(_samples[0], _samples[1])

            instantaneous_phase_difference = (
                    instantaneous_phase_difference -
                    np.mean(instantaneous_phase_difference[reference_samples, ...], axis=0))
        self.output_node.data = instantaneous_phase_difference
        # for the moment we use the rms of the entire epoch as a noise measurement
        self.output_node.rn = np.sqrt(np.mean(instantaneous_phase_difference ** 2, axis=0))

    @staticmethod
    def phase_value(v1, v2):
        vv1 = np.array([np.real(v1), np.imag(v1)]).T
        vv2 = np.array([np.real(v2), np.imag(v2)]).T
        cosine_angle = np.dot(vv1, vv2) / (np.linalg.norm(vv1) * np.linalg.norm(vv2))
        current_angle = np.arccos(cosine_angle)
        return current_angle


class InstantaneousPhaseChange(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 reference_roi_window: TimeROI = None,
                 relative_difference: bool = True,
                 reference_frequency: u.Quantity = None,
                 wavelet_half_amplitude_time: u.Quantity = 80 * u.ms,
                 **kwargs):
        """
        This InputOutputProcess computes the instantaneous phase of the data in  the inputprocess
        :param input_process: InputOutputProcess Class
        :param reference_roi_windows: reference time window to center phase
        :param relative_difference: if True, the difference using the initial phase as a reference will be returned
        :param reference_frequency: frequency to compute the relative phase difference
        :param wavelet_half_amplitude_time: length (in time) for the wavelength half-amplitude decay
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(InstantaneousPhaseChange, self).__init__(input_process=input_process, **kwargs)
        self.reference_roi_window = reference_roi_window
        self.reference_frequency = reference_frequency
        self.wavelet_half_amplitude_time = wavelet_half_amplitude_time

    def transform_data(self):
        s = (self.wavelet_half_amplitude_time.to(u.s) * self.input_node.fs.to(u.Hz))
        w = self.reference_frequency * (2 * s * np.pi) / self.input_node.fs
        y_filter = np.zeros(self.input_node.data.shape, dtype=complex)
        if self.input_node.data.ndim == 1:
            y_filter = signal.cwt(self.input_node.data.value, signal.morlet2, widths=[s], w=w)
        if self.input_node.data.ndim == 2:
            for _i in range(y_filter.shape[1]):
                y_filter[:, _i] = signal.cwt(self.input_node.data[:, _i].value, signal.morlet2,
                                             widths=[s], w=w)
        if self.input_node.data.ndim == 3:
            for _i in range(y_filter.shape[1]):
                for _j in range(y_filter.shape[2]):
                    y_filter[:, _i, _j] = signal.cwt(self.input_node.data[:, _i, _j].value, signal.morlet2,
                                                     widths=[s], w=w)
        normalized_vector = y_filter / np.abs(y_filter)
        if self.input_node.data.ndim == 3:
            normalized_vector = np.mean(normalized_vector, axis=2)
        instantaneous_phase = np.unwrap(np.angle(normalized_vector) * u.rad,
                                        axis=0)
        dy = - np.gradient(instantaneous_phase, axis=0)
        instantaneous_phase_change = - dy * (1 / self.reference_frequency) * 1 / (2 * np.pi)
        if self.reference_roi_window is not None:
            _samples = self.reference_roi_window.get_roi_samples(self.input_node)
            reference_samples = np.arange(_samples[0], _samples[1])

            instantaneous_phase_change = (
                    instantaneous_phase_change -
                    np.mean(instantaneous_phase_change[reference_samples, :], axis=0))
        self.output_node.data = instantaneous_phase_change
        # for the moment we use the rms of the entire epoch as a noise measurement
        self.output_node.rn = np.sqrt(np.mean(instantaneous_phase_change ** 2, axis=0))



"""
.. _tut-acc-standard-pipeline-sim:

##################################
ACC Standard analysis (Simulated)
##################################

In this example we simulate an ACC response and detect the peaks using n standard pipeline.

"""
# Enable below for interactive backend
# import matplotlib
# if 'Qt5Agg' in matplotlib.rcsetup.all_backends:
#    matplotlib.use('Qt5Agg')
from peegy.processing.pipe.pipeline import PipePool
from peegy.processing.pipe.simulate import GenerateInputData
from peegy.processing.tools.detection.definitions import TimePeakWindow, PeakToPeakMeasure, TimeROI
from peegy.processing.tools.template_generator.auditory_waveforms import aep
from peegy.processing.pipe.general import FilterData, ReSampling, AutoRemoveBadChannels, ReferenceData, RegressOutEOG, \
    BaselineCorrection
from peegy.processing.pipe.epochs import EpochData, AverageEpochs, ApplyFunctionToTimeROI
from peegy.processing.pipe.spatial_filtering import CreateAndApplySpatialFilter
from peegy.processing.pipe.statistics import HotellingT2Test
from peegy.processing.pipe.detection import PeakDetectionTimeDomain
from peegy.processing.pipe.plot import PlotWaveforms, PlotTopographicMap
from peegy.processing.pipe.attach import AppendGFPChannel
from peegy.processing.pipe.storage import MeasurementInformation, SubjectInformation, SaveToDatabase
import os
import numpy as np
import astropy.units as u
from peegy.processing.pipe.time_frequency import TimeFrequencyResponse, PlotTimeFrequencyData

# %%
# Time windows
# ============================
# Time windows are used to define where to look for different peaks


tw = np.array([TimePeakWindow(ini_time=50 * u.ms, end_ref='N1', label='P1', positive_peak=True,
                              exclude_channels=['GFP']),
               TimePeakWindow(ini_time=80 * u.ms, end_time=200 * u.ms, label='N1', positive_peak=False,
                              exclude_channels=['GFP']),
               TimePeakWindow(ini_ref='N1', end_time=300 * u.ms, label='P2', positive_peak=True,
                              exclude_channels=['GFP']),
               TimePeakWindow(ini_time=50 * u.ms, end_time=150 * u.ms, label='gfp1', positive_peak=True,
                              target_channels=['GFP']),
               TimePeakWindow(ini_time=150 * u.ms, end_time=500 * u.ms, label='gfp2', positive_peak=True,
                              target_channels=['GFP'])])

# %%
# Peak-to-peak Measures
# ============================
# Peak-to-peak measures are defined based on the labels of the TimePeaks defined above.


pm = np.array([PeakToPeakMeasure(ini_peak='N1', end_peak='P2')])

# %%
# Time regions of interest
# ============================
# TimeROI are defined as time regions where different measures will be performed, e.g. SNR measure or statistical
# measures


roi_windows = np.array(
    [TimeROI(ini_time=100.0 * u.ms, end_time=600.0 * u.ms, label="acc_snr"),
     TimeROI(ini_time=2100.0 * u.ms, end_time=2600.0 * u.ms, label="control")])

# %%
# Generate some data
# ===================
# First we generate some ACC data

n_channels = 32
fs = 512.0 * u.Hz
template_waveform, _ = aep(fs=fs, time_length=0.6)
event_times = np.arange(0, 400.0, 4.0) * u.s
reader = GenerateInputData(template_waveform=template_waveform,
                           fs=fs,
                           n_channels=n_channels,
                           mixing_matrix=np.diag(np.arange(n_channels))/n_channels,
                           layout_file_name='biosemi32.lay',
                           snr=0.05,
                           event_times=event_times,
                           event_code=1.0,
                           figures_subset_folder='acc_test',
                           include_eog_events=True,
                           noise_attenuation=3,
                           )
reader.run()

# %%
# Start the pipeline
# ============================
# Now we proceed with our basic processing pipeline


pipeline = PipePool()
pipeline['referenced'] = ReferenceData(reader,
                                       reference_channels=['Cz'],
                                       invert_polarity=True)
pipeline['channel_cleaned'] = AutoRemoveBadChannels(pipeline.referenced)
pipeline['down_sampled'] = ReSampling(pipeline.channel_cleaned,
                                      new_sampling_rate=256. * u.Hz)
pipeline['eog_removed'] = RegressOutEOG(pipeline.down_sampled,
                                        ref_channel_labels=['EOG1'])
pipeline.run()
# %%
# Show EOG removal Output
# ------------------------------------

pipeline.eog_removed.plot(plot_input=True,
                          plot_output=True,
                          ch_to_plot=['CP1', 'CP5', 'P7'],
                          interactive=False)

# %%
# Continue with the pipeline
# ------------------------------------

pipeline['time_filtered_data'] = FilterData(pipeline.eog_removed,
                                            high_pass=2.0 * u.Hz,
                                            low_pass=30.0 * u.Hz)

pipeline.run()

# %%
# Get Epochs
# ------------------------------------
# We partition the data into epochs or trials based on the event code used.


pipeline['time_epochs'] = EpochData(pipeline.time_filtered_data,
                                    event_code=1.0)
pipeline.run()

# %%
# Get DSS components for EOG free
# ------------------------------------
# Compute spatial filter based on EOG free epochs
pipeline['time_epochs_dss'] = CreateAndApplySpatialFilter(pipeline.time_epochs)
pipeline.run()

# %%
# Compute global field power (GFP)
# ------------------------------------
# We compute the GFP across channels and epochs. A new channel with the GFP is appended to the data

pipeline['time_epochs_with_gfp'] = AppendGFPChannel(pipeline.time_epochs)
pipeline['time_epochs_with_gfp_dss'] = AppendGFPChannel(pipeline.time_epochs_dss)
pipeline.run()

# %%
# Base line correction
# ------------------------------------
# We correct the initial segment by removing the mean in the first 20 ms
pipeline['time_epochs_with_baseline_corrected'] = BaselineCorrection(
    pipeline.time_epochs_with_gfp,
    ini_time=0 * u.ms,
    end_time=20 * u.ms)

pipeline['time_epochs_with_baseline_corrected_dss'] = BaselineCorrection(pipeline.time_epochs_with_gfp_dss,
                                                                         ini_time=0 * u.ms,
                                                                         end_time=20 * u.ms)
pipeline.run()

# %%
# Compute HT2 statistics
# ------------------------------------
# Using the raw epochs we estimate Hotelling-T2 statistics


pipeline['ht2'] = HotellingT2Test(pipeline.time_epochs_with_baseline_corrected,
                                  roi_windows=roi_windows)
pipeline['ht2_dss'] = HotellingT2Test(pipeline.time_epochs_with_baseline_corrected_dss,
                                      roi_windows=roi_windows)
pipeline.run()

# %%
# Compute average responses
# ------------------------------------
# We compute weighted average on epochs with and without spatial filtering (DSS)


pipeline['time_average_normal'] = AverageEpochs(pipeline.time_epochs_with_baseline_corrected,
                                                roi_windows=roi_windows,
                                                weighted_average=True)
pipeline['time_average_normal_gfp'] = AppendGFPChannel(pipeline.time_average_normal,
                                                       channel_label='GFP2')

pipeline['time_average_normal_dss'] = AverageEpochs(
    pipeline.time_epochs_with_baseline_corrected_dss,
    roi_windows=roi_windows,
    weighted_average=True)
pipeline.run()

# plot time-frequency response for one channel
pipeline['time_freq_transformation'] = TimeFrequencyResponse(pipeline['time_average_normal_dss'])
pipeline['plot_time_freq'] = PlotTimeFrequencyData(pipeline['time_freq_transformation'],
                                                   plot_x_lim=[0, 1],
                                                   topographic_channels=np.array(['P8']),
                                                   db_scale=False)

pipeline['roi_rms'] = ApplyFunctionToTimeROI(pipeline['time_average_normal_dss'],
                                             roi_windows=roi_windows,
                                             test_name='rms',
                                             function_object=lambda x: np.sqrt(np.mean(x ** 2, axis=0)))
pipeline['area'] = ApplyFunctionToTimeROI(
    pipeline['time_average_normal_dss'],
    roi_windows=roi_windows,
    test_name='area',
    function_object=lambda x: np.sum(np.abs(x) * 1 / pipeline['time_average_normal_dss'].output_node.fs, axis=0))
pipeline.run()
# %%
# Detect peaks
# ------------------------------------
# Using the average data, we proceed detecting peaks


pipeline['data_with_peaks'] = PeakDetectionTimeDomain(
    pipeline.time_average_normal,
    time_peak_windows=tw,
    peak_to_peak_measures=pm)

pipeline['data_with_peaks_dss'] = PeakDetectionTimeDomain(pipeline.time_average_normal_dss,
                                                          time_peak_windows=tw,
                                                          peak_to_peak_measures=pm)

# %%
# Show some waveforms
# ------------------------------------


pipeline['plotter'] = PlotWaveforms(pipeline.data_with_peaks,
                                    ch_to_plot=np.array(['C4', 'CP2', 'GFP']),
                                    overlay=[pipeline.data_with_peaks_dss],
                                    statistical_test='f_test_time',
                                    show_following_stats=['f', 'rn'],
                                    return_figures=True,
                                    )
pipeline.run()

# %%
# Show some topographic maps
# ------------------------------------


pipeline['topographic_map_1'] = PlotTopographicMap(pipeline.data_with_peaks,
                                                   topographic_channels=np.array(['C4', 'CP2', 'GFP']),
                                                   plot_x_lim=[0, 0.8],
                                                   plot_y_lim=[-3, 3],
                                                   user_naming_rule='standard',
                                                   return_figures=True)
pipeline['topographic_map_2'] = PlotTopographicMap(pipeline.data_with_peaks_dss,
                                                   topographic_channels=np.array(['C4', 'CP2', 'GFP']),
                                                   plot_x_lim=[0, 0.8],
                                                   plot_y_lim=[-3, 3],
                                                   user_naming_rule='dss',
                                                   return_figures=True)
pipeline.run()

# now we save our data to a database
subject_info = SubjectInformation(subject_id='Test_Subject')
measurement_info = MeasurementInformation(
    date='Today',
    experiment='sim')

_parameters = {'Type': 'ACC'}
database_path = reader.input_node.paths.file_directory + os.sep + 'acc_test_data.sqlite'
pipeline['database'] = SaveToDatabase(database_path=database_path,
                                      measurement_information=measurement_info,
                                      subject_information=subject_info,
                                      recording_information={'recording_device': 'dummy_device'},
                                      stimuli_information=_parameters,
                                      processes_list=[pipeline.data_with_peaks,
                                                      pipeline.data_with_peaks_dss,
                                                      pipeline.ht2,
                                                      pipeline.ht2_dss,
                                                      pipeline.roi_rms,
                                                      pipeline.area]
                                      )
pipeline.run()

# %%
# Generate pipeline diagram
# ------------------------------------
pipeline.diagram(file_name=reader.output_node.paths.figures_current_dir + 'pipeline.png',
                 return_figure=True,
                 dpi=600)



"""
This example uses an estimation of the impulse response in order to estimate and subtract the stimulus artifact.
Here we use a noise as the stimulus and its envelope plus a delay as the brain response.
"""
import numpy as np
import matplotlib.pyplot as plt
from peegy.processing.tools.epochs_processing_tools import et_impulse_response_artifact_subtraction
from peegy.processing.tools.filters.eegFiltering import bandpass_fir_win
from peegy.processing.tools.multiprocessing.multiprocessesing_filter import filt_data
from peegy.processing.tools.template_generator.auditory_waveforms import artifact_and_brain_envelope

# first we generate a test signal
fs = 16384.0
burst_duration = 1  # seconds

# here we define desired parameters for brain and artifact
stim_delay = 0.001  # neural delay in secs
brain_delay = 0.113  # brain response delay

template_waveform, stimulus_waveform, leaked_stimulus = artifact_and_brain_envelope(
    fs=fs,
    stimulus_delay=stim_delay,
    brain_delay=brain_delay,
    duration=burst_duration,
    seed=0)

# generate recorded data including brain, leaked artifact and noise
snr = -10  # dB
np.random.seed(2)
noise = 10 ** (-snr/20) * np.random.randn(template_waveform.size, 1) * np.std(template_waveform)

recorded_waveform = template_waveform + leaked_stimulus + noise

# plot stimuli, brain and leaked artifact
plt.plot(stimulus_waveform, label='input waveform')
plt.plot(recorded_waveform, label='recorded data')
plt.plot(leaked_stimulus, label='leaked artifact')
plt.plot(template_waveform, label='brain response')
plt.legend()
plt.show()

# remove artifacts
_recovered_response, _recovered_artifact = et_impulse_response_artifact_subtraction(
    data=np.atleast_3d(recorded_waveform),
    stimulus_waveform=stimulus_waveform,
    ir_length=int(0.01*fs),
    ir_max_lag=int(0.005*fs),
    regularization_factor=1,
    plot_results=True)

# apply filter to extract neural bandwidth
_b_brain = bandpass_fir_win(high_pass=2, low_pass=80.0, fs=fs)
_recovered_response = filt_data(data=_recovered_response, b=_b_brain)

# apply filter to extract neural response without removing artifact
_contaminated_response = filt_data(data=recorded_waveform, b=_b_brain)

# apply filter to extract neural response without removing artifact
_filtered_artifact = filt_data(data=leaked_stimulus, b=_b_brain)


# plot artifact
plt.figure()
plt.plot(leaked_stimulus, label='target artifact')
plt.plot(_recovered_artifact.flatten(), label='recovered artifact')
plt.legend()
plt.show()

# plot response
plt.figure()
plt.plot(_contaminated_response, label='contaminated response Corr: {:}'.format(
    np.corrcoef(template_waveform.flatten(),
                _contaminated_response[0:template_waveform.shape[0]].flatten())[-1, 0]))

plt.plot(_filtered_artifact, label='leaked artifact response Corr: {:}'.format(
    np.corrcoef(template_waveform.flatten(),
                _filtered_artifact[0:template_waveform.shape[0]].flatten())[-1, 0]))

plt.plot(template_waveform, label='target response')

plt.plot(_recovered_response.flatten(), label='recovered response. Corr: {:}'.format(
    np.corrcoef(template_waveform.flatten(),
                _recovered_response[0:template_waveform.shape[0]].flatten())[-1, 0]))
plt.legend()
plt.show()



import numpy as np
from peegy.definitions.events import SingleEvent, Events
import astropy.units as u


def detect_events(event_channel: np.array = None, fs: float = None):
    start_points = np.hstack((0, np.where(np.diff(event_channel, axis=0) != 0)[0] + 1))
    stop_points = np.hstack(
        (np.where(np.diff(event_channel, axis=0) != 0)[0] + 1, len(event_channel) - 1))
    trigger_duration = (stop_points - start_points) / fs
    events = np.array([])
    for _s, _d in zip(start_points, trigger_duration):
        events = np.append(events, SingleEvent(code=float(event_channel[round(_s)]),
                                               time_pos=_s / fs,
                                               dur=_d))
    return events


def join_trigger_events(all_events=np.array([])):
    code = np.array([])
    idx = np.array([])
    dur = np.array([])
    dur_samples = np.array([])
    event_table = {}
    for events in all_events:
        for _code, _idx, _dur, _dur_samples in zip(events['code'], events['idx'], events['dur'], events['dur_samples']):
            code = np.append(code, _code)
            idx = np.append(idx, _idx)
            dur = np.append(dur, _dur)
            dur_samples = np.append(dur_samples, _dur_samples)
        event_table = {'code': code,
                       'idx': idx,
                       'dur': dur,
                       'dur_samples': dur_samples}
    return event_table


def join_triggers(all_triggers=np.array([])):
    """
    this function join two or more sets of triggers passed in a numpy array
    :param all_triggers: all_triggers: array of triggers
    :return: a new trigger with joined elements
    """
    code = np.array([], dtype=int)
    idx = np.array([], dtype=int)
    dur = np.array([])
    dur_samples = np.array([], dtype=int)
    min_distance = np.array([], dtype=int)
    max_distance = np.array([], dtype=int)
    position = np.array([], dtype=int)

    for events in all_triggers:
        code = np.append(events['code'], code)
        idx = np.append(events['idx'], idx)
        dur = np.append(events['dur'], dur)
        dur_samples = np.append(events['dur_samples'], dur_samples)
        min_distance = np.append(events['min_distance'], min_distance)
        max_distance = np.append(events['max_distance'], max_distance)
        position = np.append(events['position'], position)

    s_idx = np.argsort(idx)
    triggers = {'code': code[s_idx],
                'idx': idx[s_idx],
                'dur': dur[s_idx],
                'dur_samples': dur_samples[s_idx],
                'min_distance': min_distance[s_idx],
                'max_distance': max_distance[s_idx],
                'position': position[s_idx],
                'triggers': []
                }
    return triggers


def get_event_times(event_channel: np.array = None, fs: u.Quantity = None):
    _events = detect_events(event_channel=event_channel, fs=fs)
    _new_events = Events(events=np.array(_events))
    print(_new_events.summary().to_string(index=False))
    return _new_events


def events_to_samples_array(events: Events = None,
                            fs: u.Quantity = None,
                            n_samples: int = None,
                            event_code: float = None,
                            min_samples_width: int = 2,
                            include_bad_events: bool = True,
                            resolve_overlap: bool = True):
    out = np.zeros((n_samples, 1))
    samples = events.get_events_index(fs=fs,
                                      include_bad_events=include_bad_events,
                                      code=event_code)
    if resolve_overlap:
        _indexes = find_repeated_values(samples=samples)
        while _indexes.size:
            _to_shift = _indexes[1::2]
            samples[_to_shift] = samples[_to_shift] + min_samples_width
            _indexes = find_repeated_values(samples=samples)

    codes = events.get_events_code(include_bad_events=include_bad_events,
                                   code=event_code)
    durations = events.get_events_duration(include_bad_events=include_bad_events)
    for _sample, _dur, _code in zip(samples, durations, codes):
        if _sample >= n_samples:
            break
        _ini = _sample
        _end = _sample + np.maximum(min_samples_width, np.floor(_dur * fs)).astype(int)
        out[_ini: _end, :] = _code
    return out


def find_repeated_values(samples: np.array = None):
    vals, inverse, count = np.unique(samples,
                                     return_inverse=True,
                                     return_counts=True,
                                     axis=0)
    _indexes = np.where(count[inverse] > 1)[0]
    return _indexes



import json
from io import open
from peegy.io.storage.data_storage_tools import MeasurementInformation
from peegy.definitions.eeg_definitions import signal_information
from peegy.io.eeg import reader
from peegy.processing.tools.detection.definitions import TimePeakWindow, PeakToPeakMeasure, TimeROI
import os
import numpy as np
from os.path import basename
from prettytable import PrettyTable
from peegy.processing.events import event_tools
import logging
from pathlib import Path
from itertools import compress
import glob
import csv
from itertools import chain
import pandas as pd
from collections.abc import MutableMapping
import astropy.units as u


def flatten_dictionary(d, parent_key='', sep='_'):
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, MutableMapping):
            items.extend(flatten_dictionary(v, new_key, sep=sep).items())
        elif isinstance(v, list):
            if np.any([isinstance(_v, MutableMapping) for _v in v]):
                for _i, _v in enumerate(v):
                    extra_sep = '' if len(v) == 1 else sep + str(_i)
                    if isinstance(_v, MutableMapping):
                        items.extend(flatten_dictionary(_v, new_key + extra_sep, sep=sep).items())
                    else:
                        items.append((new_key, _v))
            else:
                if len(v) == 1:
                    items.append((new_key, v[0]))
                else:
                    items.append((new_key, v))
        else:
            items.append((new_key, v))
    return dict(items)


class DataLinks(object):
    """
    This class is used to link the parameters associated to each data file as well as the initial and end times to
    to be read
    """

    def __init__(self, parameters: list = [],
                 parameters_file: str = None,
                 parameters_date: str = '',
                 data_file: str = None,
                 data_header: object = None,
                 data_date: str = '',
                 ini_time=0.0,
                 end_time=None,
                 meta_data: list = [],
                 label: str = None):
        self.parameters_file = parameters_file
        self.parameters_date = parameters_date
        self.data_file = data_file
        self.data_date = data_date
        self.data_header = data_header
        self.ini_time = ini_time
        self.end_time = end_time
        self.measurement_parameters = parameters
        self.meta_data = meta_data
        self.label = label


def match_keywords(key_words: [str] = None, word: str = None, deep_match=False):
    """
    This function matches a list of keywords in an string.
    :param key_words: string list of key words to be matched
    :param word: string variable where words are to be found
    :param deep_match: if True, all keywords must be present to return True
    :return: Bool variable indicating if a match was obtained
    """
    _found = False
    if deep_match:
        if key_words is not None and np.all([_key_word in word for _key_word in key_words]):
            _found = True
    else:
        if key_words is not None and any(
                [_key_word in word for _key_word in key_words]):
            _found = True
    return _found


def parse_processing_chain(file_path=''):
    with open(file_path, encoding="utf-8") as f:
        chain = json.load(f)
    all_keys = list(chain.keys())
    if 'measurements' not in all_keys:
        return
    # get default values
    defaults_list = []
    if 'defaults' in all_keys:
        defaults_list = chain['defaults']
    print(defaults_list)
    # set defaults
    out_measurements = []
    # add any default value not present in the read measurement
    template = {}
    if defaults_list:
        template = defaults_list.copy()

    for _c_measurement in chain['measurements']:
        _c_dict = template.copy()
        _signals = []
        _time_peaks = []
        _peak_to_peak_amps = []
        _measurement_info = []
        _roi_windows = []
        # add default items if they are not part of the measurement
        for _items, _val in template.items():
            if _items not in list(_c_measurement.keys()):
                _c_measurement[_items] = _val
        for _idx_item, (_field, _value) in enumerate(_c_measurement.items()):
            if _field == 'measurement_info':
                if 'measurement_info' in list(_c_dict.keys()):
                    if len(_c_dict['measurement_info']) > len(_c_measurement['measurement_info']) == 1:
                        for _c_m in _c_dict['measurement_info']:
                            _measurement_info.append(
                                MeasurementInformation(**modify_template(_c_m, _c_measurement['measurement_info'][0])))
                    elif len(_c_dict['measurement_info']) == len(_c_measurement['measurement_info']):
                        for _m_idx, _c_m in enumerate(_c_dict['measurement_info']):
                            _measurement_info.append(
                                MeasurementInformation(
                                    **modify_template(_c_m, _c_measurement['measurement_info'][_m_idx])))
                    else:
                        print('The number of measurement_info must be 1 or the same as in the template')
                else:
                    for _measure in _value:
                        _measurement_info.append(MeasurementInformation(**_measure))
                _c_dict['measurement_info'] = _measurement_info
            elif _field == 'roi_windows':
                for _roi in _value:
                    if not isinstance(_roi['measure'], list):
                        _roi['measure'] = [_roi['measure']]
                    if not isinstance(_roi['label'], list):
                        _roi['label'] = [_roi['label']]
                    assert len(_roi['label']) == len(_roi['measure'])

                    for _m, _lab in zip(_roi['measure'], _roi['label']):
                        _c_roi = _roi.copy()
                        _c_roi['measure'] = _m
                        _c_roi['label'] = _lab
                        _roi_windows.append(TimeROI(**_c_roi))
                _c_dict['roi_windows'] = _roi_windows
            elif _field == 'signal_info':
                for _signal in _value:
                    if 'signal_info' in list(_c_dict.keys()):
                        if isinstance(_c_dict['signal_info'], list):
                            for _s in _c_dict['signal_info']:
                                _signals.append(signal_information(**modify_template(_s, _signal)))
                        else:
                            _signals.append(signal_information(**modify_template(_c_dict['signal_info'], _signal)))
                    else:
                        _signals.append(signal_information(**_signal))
                _c_dict[_field] = _signals
            elif _field == 'peak_time_windows':
                for _time_peak in _value:
                    _time_peaks.append(TimePeakWindow(**_time_peak))
                _c_dict[_field] = _time_peaks
            elif _field == 'peak_to_peak_amp_labels':
                for _peak_amp in _value:
                    _peak_to_peak_amps.append(PeakToPeakMeasure(**_peak_amp))
                _c_dict[_field] = _peak_to_peak_amps
            elif _field == 'data_files':
                _data_path = os.path.dirname(_value[0])
                _c_dict['path'] = _data_path
                _c_dict[_field] = _value
            else:
                _c_dict[_field] = _value
        # convert numeric array to numpy array
        convert_dict_to_numpy(_c_dict)
        out_measurements.append(_c_dict)
    return out_measurements


def modify_template(template={}, update_dict={}):
    out_dict = template.copy()
    for _key, _value in list(update_dict.items()):
        out_dict[_key] = _value
    return out_dict


def convert_dict_to_numpy(_dict={}):
    for _field, _value in _dict.items():
        if isinstance(_value, list):
            _dict[_field] = np.array(_value)
        if isinstance(_value, dict):
            convert_dict_to_numpy(_dict[_field])


def get_files_and_meta_files(measurement_path: str = '',
                             filter_key_words_in_path: [str] = None,
                             exclude_key_words_in_path: [str] = None,
                             filter_key_words_in_file: [str] = None,
                             exclude_key_words_in_file: [str] = None,
                             deep_match_in_filter_path: bool = True,
                             deep_match_in_filter_file: bool = True,
                             deep_match_in_exclude_path: bool = True,
                             deep_match_in_exclude_file: bool = True,
                             ignore_meta_file=False,
                             file_types: [str] = ['bdf', 'edf'],
                             meta_file_extension: str = 'json',
                             split_trigger_codes: [float] = None,
                             ) -> pd.DataFrame:
    """
    This function search for files which extension is defined by file_types and meta_files associated to each file.
    Each file will be associated to the meta_file in the output dataframe.
    This is, in order to pair a bdf to json file both should have the same name.
    The result is a list of DataLinks object, containing the information that links both files.
    If the meta_file_type is a json file, the json file will be parsed as a dictionary in the DataLinks parameters and
    also in the output dataframe. In this way, the data to be processed can be manipulated using pandas
    functionalities.
    :param measurement_path: the root path to search for pairs of files
    :param filter_key_words_in_path: if provided, the return list will only contain DataLinks object whose paths does
    include
    the exclude_key_words_in_path. This is useful to process entire folders with different conditions
    :param exclude_key_words_in_path: if provided, the return list will only contain DataLinks object whose paths does
     not include
    the exclude_key_words_in_path. This is useful to process entire folders with different conditions
    :param filter_key_words_in_file: if provided, the return list will only contain DataLinks object whose file name
     does include
    the filter_key_words_in_file. This is useful to process entire folders with different conditions
    :param exclude_key_words_in_file: if provided, the return list will only contain DataLinks object whose file name
    does not include
    the exclude_key_words_in_path. This is useful to process entire folders with different conditions
    :param deep_match_in_filter_path: if True, all keywords must be present to return True
    :param deep_match_in_filter_file: if True, all keywords must be present to return True
    :param deep_match_in_exclude_path: if True, all keywords must be present to return True
    :param deep_match_in_exclude_file: if True, all keywords must be present to return True
    :param meta_file_extension: if True it will not care whether json file is present or not
    :param file_types: a list of strings indicating the file types to be found, e.g. .bdf, .edf, .txt, .json, etcetera
    :param split_trigger_codes: list of trigger codes. If provided, the bdf file will be marked in chunks starting at
    each of these trigger events.
    The time stamps indicating the beginning and end of eeach block will be passed in  time the ini_time and
    end_time of the DataLinks class.
    :return: a Pandas data frame with DataLinks objects which contain the information linking the files
    """
    data_files = []
    for files in file_types:
        data_files.extend(Path(measurement_path).glob('**/*.{:}'.format(files)))

    # avoid hidden folders
    _to_keep = []
    for _file in data_files:
        _to_keep.append(any([_parts.startswith('.') for _parts in os.path.normpath(str(_file)).split(os.sep)]) is False)
    data_files = list(compress(data_files, _to_keep))

    par_out = pd.DataFrame()
    for i, _full_file_name in enumerate(data_files):
        _data_file_name, _file_extension = os.path.splitext(_full_file_name)
        _meta_file_path = ''
        if meta_file_extension:
            _meta_file_path = _data_file_name + '.{:}'.format(meta_file_extension)
        if filter_key_words_in_path is not None and not match_keywords(
                filter_key_words_in_path, str(_full_file_name),
                deep_match=deep_match_in_filter_path):
            continue
        if exclude_key_words_in_path is not None and match_keywords(
                exclude_key_words_in_path, str(_full_file_name),
                deep_match=deep_match_in_exclude_path):
            continue

        if filter_key_words_in_file is not None and \
                not (match_keywords(filter_key_words_in_file, _data_file_name, deep_match=deep_match_in_filter_file) and
                     match_keywords(filter_key_words_in_file, _meta_file_path, deep_match=deep_match_in_filter_file)):
            continue
        if exclude_key_words_in_file is not None and \
                (match_keywords(exclude_key_words_in_file, _data_file_name, deep_match=deep_match_in_exclude_file) or
                 match_keywords(exclude_key_words_in_file, _meta_file_path, deep_match=deep_match_in_exclude_file)):
            continue

        _parameters = {}

        if not ignore_meta_file and os.path.isfile(_meta_file_path) and meta_file_extension == 'json':
            with open(_meta_file_path, 'r') as f:
                try:
                    _parameters = json.load(f)
                except ValueError:
                    print('could not open {:}'.format(_meta_file_path))

        if split_trigger_codes is not None:
            _data = reader.eeg_reader(_full_file_name)
            raw_events = _data.get_events()
            events = event_tools.detect_events(event_channel=raw_events, fs=_data.fs)
            print(event_tools.Events(events=events).summary())
            _split_events = np.array([_e for _e in events if _e.code in split_trigger_codes])
            ini_time = [_e.time_pos for _e in _split_events]
            end_time = [_e.time_pos for _e in _split_events[1:]]
            end_time.append(None)
            split_codes = [_e.code for _e in _split_events]
        else:
            ini_time = [0 * u.s]
            end_time = [None]
            split_codes = [None]

        for _ini_time, _end_time, _code in zip(ini_time, end_time, split_codes):
            data_links = DataLinks(parameters=_parameters,
                                   parameters_file=_meta_file_path,
                                   data_file=_full_file_name,
                                   ini_time=_ini_time,
                                   end_time=_end_time,
                                   label=_code
                                   )

            par_out = pd.concat([par_out, pd.DataFrame.from_dict(
                dict({k: [v] for k, v in _parameters.items()},
                     **{'data_links': [data_links]}))],
                                ignore_index=True)
    if par_out.size:
        t = PrettyTable()
        _aux = par_out.data_links.values.astype(np.ndarray)
        t.add_column(fieldname='File name', column=[basename(x.parameters_file) for x in _aux])
        t.add_column(fieldname='File date', column=[basename(x.parameters_date) for x in _aux])
        t.add_column(fieldname='data file name', column=[basename(x.data_file) for x in _aux])
        t.add_column(fieldname='data file date', column=[x.data_date for x in _aux])
        t.add_column(fieldname='Code', column=[x.label for x in _aux])
        t.add_column(fieldname='Ini time', column=[x.ini_time for x in _aux])
        t.add_column(fieldname='End time', column=[x.end_time for x in _aux])
        logging.info(t)
        print(t)
    else:
        print('No files were found')
    return par_out


def get_data_and_meta_data_files(measurement_path: str = '',
                                 filter_key_words_in_path: [str] = None,
                                 exclude_key_words_in_path: [str] = None,
                                 filter_key_words_in_file: [str] = None,
                                 exclude_key_words_in_file: [str] = None,
                                 deep_match_in_filter_path: bool = True,
                                 deep_match_in_filter_file: bool = True,
                                 deep_match_in_exclude_path: bool = True,
                                 deep_match_in_exclude_file: bool = True,
                                 meta_data_file_extension: [str] = ['.json'],
                                 rows_to_skip=0,
                                 ) -> [DataLinks]:
    """
    This function search for .bdf or .edf  and .json files recursively and pair them according to name.
    This is, in order to pair a bdf to json file both should have at least the same name (it may be longer).
    The result is a list of DataLinks object, containing the information that links both files.
    :param measurement_path: the root path to search for pairs of files
    :param filter_key_words_in_path: if provided, the return list will only contain DataLinks object whose paths does
     include
    the exclude_key_words_in_path. This is useful to process entire folders with different conditions
    :param exclude_key_words_in_path: if provided, the return list will only contain DataLinks object whose paths does
    not include
    the exclude_key_words_in_path. This is useful to process entire folders with different conditions
    :param filter_key_words_in_file: if provided, the return list will only contain DataLinks object whose file name
    does include
    the filter_key_words_in_file. This is useful to process entire folders with different conditions
    :param exclude_key_words_in_file: if provided, the return list will only contain DataLinks object whose file name
    does not include
    the exclude_key_words_in_path. This is useful to process entire folders with different conditions
    :param deep_match_in_filter_path: if True, all keywords must be present to return True
    :param deep_match_in_filter_file: if True, all keywords must be present to return True
    :param deep_match_in_exclude_path: if True, all keywords must be present to return True
    :param deep_match_in_exclude_file: if True, all keywords must be present to return True
    :param meta_data_file_extension: list indicating the extension of associated meta_data files, e.g. ['.csv', .'json']
    :param rows_to_skip: number of rows to skip when reading csv file
    :return: a list of DataLinks objects which contain the information linking .bdf or .edf and .json files
    """
    types = ('**/*.bdf', '**/*.edf')  # the tuple of file types
    data_files = []
    for files in types:
        data_files.extend(Path(measurement_path).glob(files))

    par_out = []
    for i, _full_file_name in enumerate(data_files):
        _data_file_name, _file_extension = os.path.splitext(_full_file_name)
        if filter_key_words_in_path is not None and not match_keywords(
                filter_key_words_in_path, str(_full_file_name),
                deep_match=deep_match_in_filter_path):
            continue
        if exclude_key_words_in_path is not None and match_keywords(
                exclude_key_words_in_path, str(_full_file_name),
                deep_match=deep_match_in_exclude_path):
            continue

        if filter_key_words_in_file is not None and not (
                match_keywords(filter_key_words_in_file, _data_file_name, deep_match=deep_match_in_filter_file)):
            continue
        if exclude_key_words_in_file is not None and (
                match_keywords(exclude_key_words_in_file, _data_file_name, deep_match=deep_match_in_exclude_file)):
            continue

        meta_files = list(chain(*[glob.glob(_data_file_name + '*{:}'.format(_extension))
                                  for _extension in meta_data_file_extension]))
        meta_data = []
        for _meta_file in meta_files:
            _meta_file_name, file_extension = os.path.splitext(_meta_file)
            if filter_key_words_in_file is not None and not (
                    match_keywords(filter_key_words_in_file, _meta_file_name, deep_match=deep_match_in_filter_file)):
                continue
            if exclude_key_words_in_file is not None and (
                    match_keywords(exclude_key_words_in_file, _meta_file_name, deep_match=deep_match_in_exclude_file)):
                continue
            if file_extension == '.json':
                with open(_meta_file, 'r') as f:
                    meta_data.append(json.load(f))
            columns = []
            if file_extension == '.csv':
                with open(_meta_file, 'r') as f:
                    reader = csv.reader(f)
                    for _i_r, row in enumerate(reader):
                        if _i_r < rows_to_skip:
                            continue
                        if _i_r == rows_to_skip:
                            # first row
                            columns_names = [[value] for value in row]
                            columns = [[np.array([])] * len(columns_names)]
                            continue

                        for _c, value in enumerate(row):
                            columns[0][_c] = np.append(columns[0][_c], value)
                    current_meta_data = dict()
                    for _col_idx, _col_name in enumerate(columns_names):
                        current_meta_data[_col_name[0]] = columns[0][_col_idx]
                    meta_data.append(current_meta_data)

        data_links = DataLinks(parameters=meta_data,
                               parameters_file=_meta_file,
                               data_file=_full_file_name,
                               )
        par_out.append(data_links)

    t = PrettyTable()
    t.add_column(fieldname='File name', column=[basename(x.parameters_file) for x in par_out])
    t.add_column(fieldname='File date', column=[basename(x.parameters_date) for x in par_out])
    t.add_column(fieldname='bdf file name', column=[basename(x.data_file) for x in par_out])
    t.add_column(fieldname='bdf file date', column=[x.data_date for x in par_out])
    logging.info(t)
    print(t)
    return par_out
